import numpy as np
import numpy.testing as npt

from ...coordinates.origin import Origin
from ..kepler import (
    _calc_elliptical_anomalies,
    _calc_hyperbolic_anomalies,
    _calc_parabolic_anomalies,
    calc_apoapsis_distance,
    calc_mean_anomaly,
    calc_mean_motion,
    calc_periapsis_distance,
    calc_period,
    calc_semi_latus_rectum,
    calc_semi_major_axis,
    solve_kepler,
)

RELATIVE_TOLERANCE = 0.0
ABSOLUTE_TOLERANCE = 1e-15

# --- Tests last updated: 2023-12-18


def test_calc_mean_anomaly_elliptical():
    # Test mean anomaly calculations for elliptical orbits
    # At the two apses, mean anomaly and true anomaly
    # should be equal.

    # Periapsis point
    nu = 0.0
    e = 0.5
    M_desired = nu
    M_actual = calc_mean_anomaly(nu, e)
    npt.assert_allclose(
        M_actual, M_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )

    # Apoapsis point
    nu = np.pi
    e = 0.5
    M_desired = nu
    M_actual = calc_mean_anomaly(nu, e)
    npt.assert_allclose(
        M_actual, M_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )

    # Periapsis point (at 360 degrees)
    nu = 2 * np.pi
    e = 0.5
    M_desired = 0.0
    M_actual = calc_mean_anomaly(nu, e)
    npt.assert_allclose(
        M_actual, M_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )


def test__calc_elliptical_anomalies():
    # Test eccentric anomaly and mean anomaly calculations
    # for elliptical orbits. At the two apses, mean anomaly,
    # eccentric anomaly and true anomaly should be equal.

    # Periapsis point
    nu = 0.0
    e = 0.5
    M_desired = nu
    E_desired = nu
    E_actual, M_actual = _calc_elliptical_anomalies(nu, e)
    npt.assert_allclose(
        (E_actual, M_actual),
        (E_desired, M_desired),
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )

    # Apoapsis point
    nu = np.pi
    e = 0.5
    M_desired = nu
    E_desired = nu
    E_actual, M_actual = _calc_elliptical_anomalies(nu, e)
    npt.assert_allclose(
        (E_actual, M_actual),
        (E_desired, M_desired),
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )

    # Periapsis point (at 360 degrees)
    nu = 2 * np.pi
    e = 0.5
    M_desired = 0.0
    E_desired = 0.0
    E_actual, M_actual = _calc_elliptical_anomalies(nu, e)
    npt.assert_allclose(
        (E_actual, M_actual),
        (E_desired, M_desired),
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )


def test__calc_parabolic_anomalies():
    # Test eccentric anomaly and mean anomaly calculations
    # for parabolic orbits.  At the only valid apsis, mean anomaly,
    # eccentric anomaly and true anomaly should be equal.

    # Periapsis point
    nu = 0.0
    e = 1.0
    M_desired = nu
    E_desired = nu
    E_actual, M_actual = _calc_parabolic_anomalies(nu, e)
    npt.assert_allclose(
        (E_actual, M_actual),
        (E_desired, M_desired),
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )


def test__calc_hyperbolic_anomalies():
    # Test eccentric anomaly and mean anomaly calculations
    # for hyperbolic orbits. At the only valid apsis, mean anomaly,
    # eccentric anomaly and true anomaly should be equal.

    # Periapsis point
    nu = 0.0
    e = 1.5
    M_desired = nu
    E_desired = nu
    E_actual, M_actual = _calc_hyperbolic_anomalies(nu, e)
    npt.assert_allclose(
        (E_actual, M_actual),
        (E_desired, M_desired),
        rtol=RELATIVE_TOLERANCE,
        atol=ABSOLUTE_TOLERANCE,
    )


def test_solve_kepler_elliptical():
    # Test true anomaly  calculations for elliptical orbits.
    # At the two apses, true anomaly and mean anomaly
    # should be equal.

    # Periapsis point
    e = 0.5
    M = 0.0
    nu_desired = 0.0
    nu_actual = solve_kepler(e, M)
    npt.assert_allclose(
        nu_actual, nu_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )

    # Apoapsis point
    e = 0.5
    M = np.pi
    nu_desired = np.pi
    nu_actual = solve_kepler(e, M)
    npt.assert_allclose(
        nu_actual, nu_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )

    # Periapsis point (at 360 degrees)
    e = 0.5
    M = 2 * np.pi
    nu_desired = 0.0
    nu_actual = solve_kepler(e, M)
    npt.assert_allclose(
        nu_actual, nu_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )


def test_solve_kepler_parabolic():
    # Test true anomaly calculations for parabolic orbits.
    # At the only valid apsis, true anomaly and mean anomaly
    # should be equal.

    # Periapsis point
    e = 1.0
    M = 0.0
    nu_desired = 0.0
    nu_actual = solve_kepler(e, M)
    npt.assert_allclose(
        nu_actual, nu_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )


def test_solve_kepler_hyperbolic():
    # Test true anomaly calculations for hyperbolic orbits.
    # At the only valid apsis, true anomaly and mean anomaly
    # should be equal.

    # Periapsis point
    e = 1.5
    M = 0.0
    nu_desired = 0.0
    nu_actual = solve_kepler(e, M)
    npt.assert_allclose(
        nu_actual, nu_desired, rtol=RELATIVE_TOLERANCE, atol=ABSOLUTE_TOLERANCE
    )


def test_solve_kepler_elliptical_orbits(orbital_elements):
    # Test true anomaly calculations for elliptical orbits.
    # Limit orbits to those with eccentricity < 1.0
    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    nu_desired = orbital_elements["nu"]
    M = orbital_elements["M"].values
    e = orbital_elements["e"].values

    for e_i, M_i, nu_desired_i in zip(e, M, nu_desired):
        nu_actual = np.degrees(solve_kepler(e_i, np.radians(M_i)))
        npt.assert_allclose(nu_actual, nu_desired_i, rtol=0.0, atol=1e-12)


def test_solve_kepler_hyperbolic_orbits(orbital_elements):
    # Test true anomaly calculations for elliptical orbits.
    # Limit orbits to those with eccentricity > 1.0
    orbital_elements = orbital_elements[orbital_elements["e"] > 1.0]
    nu_desired = orbital_elements["nu"]
    M = orbital_elements["M"].values
    e = orbital_elements["e"].values

    for e_i, M_i, nu_desired_i in zip(e, M, nu_desired):
        nu_actual = np.degrees(solve_kepler(e_i, np.radians(M_i)))
        npt.assert_allclose(nu_actual, nu_desired_i, rtol=0.0, atol=1e-12)


def test_calc_period(orbital_elements):
    # Test period calculations for elliptical orbits
    P_desired = orbital_elements["P"]
    # For parabolic/hyperbolic orbits, orbital period is infinite
    # Our test data from JPL Horizons uses a value of 1e99 for infinite
    P_desired = np.where(P_desired > 1e99, np.inf, P_desired)
    a = orbital_elements["a"].values
    origin = Origin.from_kwargs(
        code=np.full(len(orbital_elements), "SUN", dtype="object")
    )
    mu = origin.mu()

    P_actual = calc_period(a, mu)
    npt.assert_allclose(P_actual, P_desired, rtol=0.0, atol=1e-10)


def test_calc_periapsis_distance(orbital_elements):
    # Test periapse distance calculations
    q_desired = orbital_elements["q"]
    a = orbital_elements["a"].values
    e = orbital_elements["e"].values

    q_actual = calc_periapsis_distance(a, e)
    npt.assert_allclose(q_actual, q_desired, rtol=0.0, atol=1e-12)


def test_calc_apoapsis_distance(orbital_elements):
    # Test apoapsis distance calculations
    Q_desired = orbital_elements["Q"]
    # For parabolic/hyperbolic orbits, apoapsis distance is infinite
    # Our test data from JPL Horizons uses a value of 1e99 for infinite
    Q_desired = np.where(Q_desired > 1e99, np.inf, Q_desired)
    a = orbital_elements["a"].values
    e = orbital_elements["e"].values

    Q_actual = calc_apoapsis_distance(a, e)
    npt.assert_allclose(Q_actual, Q_desired, rtol=0.0, atol=1e-12)


def test_calc_semi_major_axis(orbital_elements):
    # Test semi-major axis calculations
    a_desired = orbital_elements["a"]
    q = orbital_elements["q"].values
    e = orbital_elements["e"].values

    a_actual = calc_semi_major_axis(q, e)
    npt.assert_allclose(a_actual, a_desired, rtol=0.0, atol=1e-12)


def test_calc_semi_latus_rectum(orbital_elements):
    # Test semi-latus rectum calculations
    a = orbital_elements["a"].values
    e = orbital_elements["e"].values
    p_desired = (1 - e**2) * a

    p_actual = calc_semi_latus_rectum(a, e)
    npt.assert_allclose(p_actual, p_desired, rtol=0.0, atol=1e-12)


def test_calc_mean_motion(orbital_elements):
    # Test mean motion calculations
    n_desired = orbital_elements["n"]
    a = orbital_elements["a"].values
    origin = Origin.from_kwargs(
        code=np.full(len(orbital_elements), "SUN", dtype="object")
    )
    mu = origin.mu()

    n_actual = np.degrees(calc_mean_motion(a, mu))
    npt.assert_allclose(n_actual, n_desired, rtol=0.0, atol=1e-12)



from __future__ import annotations

from typing import Iterator

import pyarrow.compute as pc
import quivr as qv

from .detections import PointSourceDetections


class Associations(qv.Table):

    detection_id = qv.LargeStringColumn()
    object_id = qv.LargeStringColumn(nullable=True)

    # TODO: We may want to create a derivative class called "ProbabilisticAssociations" that
    # includes residuals with respect to an orbit
    # orbit_id = qv.LargeStringColumn(nullable=True)
    # residuals = Residuals.as_column(nullable=True) # from adam_core.coordinates.residuals import Residuals

    def group_by_object(self) -> Iterator["Associations"]:
        """
        Group the associations by object ID. Any null object IDs will be returned last.

        Returns
        -------
        associations : Iterator[`~adam_core.observations.associations.Associations`]
            Associations grouped by object ID.
        """
        # Gather unique exposure IDs
        object_ids = self.object_id.unique()
        sorted = self.table.sort_by("object_id")

        # Return non-null object IDs first
        for object_id in pc.drop_null(object_ids):
            mask = pc.equal(sorted.column("object_id"), object_id)
            table = sorted.filter(mask)
            yield Associations.from_pyarrow(table)

        # If there are any null object IDs, return them last
        if object_ids.null_count > 0:
            mask = pc.is_null(sorted.column("object_id"))
            table = sorted.filter(mask)
            yield Associations.from_pyarrow(table)

    def link_to_detections(
        self, detections: PointSourceDetections
    ) -> qv.Linkage[Associations, PointSourceDetections]:
        """
        Link the associations to a table of detections.

        Parameters
        ----------
        detections : `~adam_core.observations.detections.PointSourceDetections`
            Table of detections to link to.
        """
        # NOTE: Unsure if this is actually useful since its 1 to 1
        return qv.Linkage(self, detections, self.detection_id, detections.id)



import logging
from typing import Optional

import ray

logger = logging.getLogger(__name__)


def initialize_use_ray(
    num_cpus: Optional[int] = None, object_store_bytes: Optional[int] = None, **kwargs
) -> bool:
    """
    Ensures we use existing local cluster, or starts new one with desired resources
    """
    use_ray = False
    if num_cpus is None or num_cpus > 1:
        # Initialize ray
        if not ray.is_initialized():
            logger.info("Ray is not initialized. Initializing...")
            # For some reason, ray does not seem to automatically
            # find existing local clusters without `address="auto"`
            # but it will fail if we use auto and there is no existing cluster.
            # So we wrap it in a try/except, using an existing cluster if we can
            # Otherwise starting fresh.
            try:
                logger.info("Attempting to connect to existing ray cluster...")
                ray.init(address="auto")
            except ConnectionError:
                logger.info("Could not connect to existing ray cluster.")
                logger.info(
                    f"Attempting ray with {num_cpus} cpus and {object_store_bytes} bytes."
                )
                ray.init(
                    dashboard_host="0.0.0.0",
                    num_cpus=num_cpus,
                    object_store_memory=object_store_bytes,
                    **kwargs,
                )

        logger.info(f"Ray Resources: {ray.cluster_resources()}")

        use_ray = True
    return use_ray



from typing import Iterable, Sequence, Tuple

import numpy as np

MILLISECOND_IN_DAYS = 1 / 86400 / 1000


def _iterate_chunks(iterable: Sequence, chunk_size: int) -> Iterable:
    """
    Generator that yields chunks of size chunk_size from sized iterable
    (Sequence).

    Parameters
    ----------
    iterable : Sequence
        Iterable to chunk.
    chunk_size : int
        Size of chunks.

    Yields
    ------
    chunk : iterable
        Chunk of size chunk_size from iterable.
    """
    N = len(iterable)
    for i in range(0, N, chunk_size):
        yield iterable[i : i + chunk_size]


def _iterate_chunk_indices(
    iterable: Sequence, chunk_size: int
) -> Iterable[Tuple[int, int]]:
    """
    Generator that yields indices for chunks of size chunk_size from sized
    iterable (Sequence).

    Parameters
    ----------
    iterable : Sequence
        Iterable to chunk.
    chunk_size : int
        Size of chunks.

    Yields
    ------
    chunk : iterable
        Chunk of size chunk_size from iterable.
    """
    N = len(iterable)
    for i in range(0, N, chunk_size):
        yield i, min(i + chunk_size, N)


def _assert_times_almost_equal(
    have: np.ndarray, want: np.ndarray, tolerance: float = 0.1
):
    """
    Raises a ValueError if the time arrays (in units of days such as MJD) are not within the
    tolerance in milliseconds of each other.

    Parameters
    ----------
    have : `~numpy.ndarray`
        Times (in units of days) to check.
    want : `~numpy.ndarray`
        Times (in units of days) to check.

    Raises
    ------
    ValueError: If the time arrays are not within the tolerance in milliseconds of each other.
    """
    tolerance_in_days = tolerance * MILLISECOND_IN_DAYS

    diff = np.abs(have - want)
    if np.any(diff > tolerance_in_days):
        raise ValueError(f"Times were not within {tolerance:.6f} ms of each other.")



import warnings
from typing import List, Tuple, Union

import numpy as np
import numpy.typing as npt
import pyarrow as pa
import quivr as qv
from scipy import stats

from .cartesian import CartesianCoordinates
from .cometary import CometaryCoordinates
from .keplerian import KeplerianCoordinates
from .spherical import SphericalCoordinates

CoordinateType = Union[
    CartesianCoordinates,
    CometaryCoordinates,
    KeplerianCoordinates,
    SphericalCoordinates,
]
SUPPORTED_COORDINATES = (
    CartesianCoordinates,
    CometaryCoordinates,
    KeplerianCoordinates,
    SphericalCoordinates,
)

__all__ = [
    "Residuals",
    "calculate_chi2",
    "_batch_coords_and_covariances",
]


def apply_cosine_latitude_correction(
    lat: npt.NDArray[np.float64],
    residuals: npt.NDArray[np.float64],
    covariances: npt.NDArray[np.float64],
) -> Tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Apply a correction factor of cosine latitude to the residuals and covariance matrix.
    This is designed to account for the fact that longitudes get smaller as you approach
    the poles.

    Parameters
    ----------
    lat : `~numpy.ndarray` (N)
        Latitudes in degrees.
    residuals : `~numpy.ndarray` (N, D)
        Spherical residuals.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices for spherical residuals.

    Returns
    -------
    residuals : `~numpy.ndarray` (N, D)
        Spherical residuals with the correction factor applied.
    covariances : `~numpy.ndarray` (N, D, D)
        Covariance matrices for spherical residuals with the correction factor applied.
    """
    N = len(lat)
    cos_lat = np.cos(np.radians(lat))

    identity = np.identity(6, dtype=np.float64)

    # Populate the diagonal of the matrix with cos(latitude) for
    # the longitude and longitudinal velocity dimensions
    diag = np.ones((N, 6))
    diag[:, 1] = cos_lat
    diag[:, 4] = cos_lat

    # Calculate the cos(latitude) factor for the covariance matrix
    cos_lat_cov = np.einsum("kj,ji->kij", diag, identity, order="C")

    # Apply the cos(latitude) factor to the residuals in longitude
    # and longitudinal velocity
    residuals[:, 1] *= cos_lat
    residuals[:, 4] *= cos_lat

    # Identify locations where the covariance matrix is NaN and set
    # those values to 0.0
    nan_cov = np.isnan(covariances)
    covariances_masked = np.where(nan_cov, 0.0, covariances)

    # Apply the cos(latitude) factor to the covariance matrix.
    covariances_masked = (
        cos_lat_cov @ covariances_masked @ cos_lat_cov.transpose(0, 2, 1)
    )

    # Set the covariance matrix to nan where it was nan before.
    # Note that when we apply cos(latitude) to the covariance matrix in the previous statement,
    # we are only modifying the longitude and longitudinal velocity dimensions. The remaining
    # dimensions are left unchanged which is why we can use a mask to reset the
    # the missing values in the resulting matrix to NaN.
    covariances = np.where(nan_cov, np.nan, covariances_masked)
    return residuals, covariances


def bound_longitude_residuals(
    observed: npt.NDArray[np.float64], residuals: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Bound the longitude residuals to the range [-180, 180] degrees and adjust the
    signs of the residuals that cross the 0/360 degree boundary. By convention, we define
    residuals that cross from <360 to >0 as a positive residual (355 - 5 = +10) and cross
    from >0 to <360 as a negative residual (5 - 355 = -10).

    Parameters
    ----------
    observed : `~numpy.ndarray` (N, D)
        Observed coordinates in degrees. We use the observed coordinates to
        determine whether the residuals cross the 0/360 degree boundary.
    residuals : `~numpy.ndarray` (N, D)
        Residuals in degrees.

    Returns
    -------
    residuals : `~numpy.ndarray` (N, D)
        Residuals wrapped to the range [-180, 180] degrees.
    """
    # Extract the observed longitude and residuals
    observed_longitude = observed[:, 1]
    longitude_residual = residuals[:, 1]

    # Identify residuals that exceed a half circle
    longitude_residual_g180 = longitude_residual > 180
    longitude_residual_l180 = longitude_residual < -180

    # Wrap the residuals to the range [-180, 180] degrees
    longitude_residual = np.where(
        longitude_residual_g180, longitude_residual - 360, longitude_residual
    )
    longitude_residual = np.where(
        longitude_residual_l180, longitude_residual + 360, longitude_residual
    )

    # Adjust the signs of the residuals that cross the 0/360 degree boundary
    # We want it so that crossing from <360 to >0 is a positive residual (355 - 5 = +10)
    # and crossing from >0 to <360 is a negative residual (5 - 355 = -10)
    longitude_residual = np.where(
        longitude_residual_g180 & (observed_longitude > 180),
        -longitude_residual,
        longitude_residual,
    )
    longitude_residual = np.where(
        longitude_residual_l180 & (observed_longitude < 180),
        -longitude_residual,
        longitude_residual,
    )

    residuals[:, 1] = longitude_residual
    return residuals


class Residuals(qv.Table):

    values = qv.ListColumn(pa.float64(), nullable=True)
    chi2 = qv.Float64Column(nullable=True)
    dof = qv.Int64Column(nullable=True)
    probability = qv.Float64Column(nullable=True)

    @classmethod
    def calculate(
        cls, observed: CoordinateType, predicted: CoordinateType
    ) -> "Residuals":
        """
        Calculate the residuals between the observed and predicted coordinates. Residuals
        are defined as the observed coordinates minus the predicted coordinates. The observed
        coordinate's covariance matrix is used to calculate the chi2 and degrees of freedom.

        TODO: We may want to add support for a single predicted coordinate and covariance
            compared to multiple observed coordinates and covariances.
            Add support for cases where both the observed and predicted coordinates have
            covariances. Maybe a Kolmogorov-Smirnov test?

        Parameters
        ----------
        observed : CoordinateType (N, D)
            Observed coordinates.
        predicted : CoordinateType (N, D) or (1, D)
            Predicted coordinates. If a single coordinate is provided, it is broadcasted
            to the same shape as the observed coordinates.

        Returns
        -------
        residuals : `~adam_core.coordinates.residuals.Residuals`
            Residuals between the observed and predicted coordinates.
        """
        if not isinstance(observed, SUPPORTED_COORDINATES):
            raise TypeError(
                f"Observed coordinates must be one of {SUPPORTED_COORDINATES}, not {type(observed)}."
            )
        if not isinstance(predicted, SUPPORTED_COORDINATES):
            raise TypeError(
                f"Predicted coordinates must be one of {SUPPORTED_COORDINATES}, not {type(predicted)}."
            )
        if type(observed) != type(predicted):
            raise TypeError(
                "Observed and predicted coordinates must be the same type, "
                f"not {type(observed)} and {type(predicted)}."
            )
        if (observed.origin != predicted.origin).all():
            raise ValueError(
                "Observed and predicted coordinates must have the same origin."
            )
        if observed.frame != predicted.frame:
            raise ValueError(
                f"Observed ({observed.frame}) and predicted ({predicted.frame}) coordinates must have the same frame."
            )
        # Extract the observed and predicted values
        observed_values = observed.values
        observed_covariances = observed.covariance.to_matrix()
        predicted_values = predicted.values

        # Create the output arrays
        N, D = observed_values.shape
        p = np.empty(N, dtype=np.float64)
        chi2s = np.empty(N, dtype=np.float64)

        # Calculate the degrees of freedom for every coordinate
        # Number of coordinate dimensions less the number of quantities that are NaN
        dof = D - np.sum(np.isnan(observed_values), axis=1)

        # Calculate the array of residuals
        residuals = observed_values - predicted_values

        # If the coordinates are spherical then we do some extra work:
        # 1. Bound residuals in longitude to the range [-180, 180] degrees and
        #    adjust the signs of the residuals that cross the 0/360 degree boundary.
        # 2. Apply the cos(latitude) factor to the residuals in longitude and longitudinal
        #    velocity
        if isinstance(observed, SphericalCoordinates):
            # Bound the longitude residuals to the range [-180, 180] degrees
            residuals = bound_longitude_residuals(observed_values, residuals)

            # Apply the cos(latitude) factor to the residuals and covariance matrix
            residuals, observed_covariances = apply_cosine_latitude_correction(
                observed_values[:, 2], residuals, observed_covariances
            )

        # Batch the coordinates and covariances into groups that have the same
        # number of dimensions that have missing values (represented by NaNs)
        (
            batch_indices,
            batch_dimensions,
            batch_coords,
            batch_covariances,
        ) = _batch_coords_and_covariances(observed_values, observed_covariances)

        for indices, dimensions, coords, covariances in zip(
            batch_indices, batch_dimensions, batch_coords, batch_covariances
        ):
            if not np.all(np.isnan(covariances)):
                # Filter residuals by dimensions that have values
                residuals_i = residuals[:, dimensions]

                # Then filter by rows that belong to this batch
                residuals_i = residuals_i[indices, :]

                # Calculate the chi2 for each coordinate (this is actually
                # calculating mahalanobis distance squared -- both are equivalent
                # when the covariance matrix is diagonal, mahalanobis distance is more
                # general as it allows for covariates between dimensions)
                chi2_values = calculate_chi2(residuals_i, covariances)

                # For a normally distributed random variable, the mahalanobis distance
                # squared in D dimesions follows a chi2 distribution with D degrees of freedom.
                # So for each coordinate, calculate the probability that you would
                # get a chi2 value greater than or equal to that coordinate's chi2 value.
                # At a residual of zero this probability is 1.0, and at a residual of
                # 1 sigma (for 1 degree of freedom) this probability is ~0.3173.
                p[indices] = 1 - stats.chi2.cdf(chi2_values, dof[indices])

                # Set the chi2 for each coordinate
                chi2s[indices] = chi2_values

            else:
                # If the covariance matrix is all NaNs, then the chi2 is NaN
                chi2s[indices] = np.nan
                p[indices] = np.nan

        return cls.from_kwargs(
            values=residuals.tolist(),
            chi2=chi2s,
            dof=dof,
            probability=p,
        )

    def to_array(self) -> npt.NDArray[np.float64]:
        """
        Convert the residuals to a numpy array.

        Returns
        -------
        residuals : `~numpy.ndarray` (N, D)
            Array of residuals.
        """
        return np.stack(self.values.to_numpy(zero_copy_only=False))


def calculate_chi2(
    residuals: npt.NDArray[np.float64], covariances: npt.NDArray[np.float64]
) -> npt.NDArray[np.float64]:
    """
    Calculate the vectorized normalized squared residual for each residual and covariance pair.
    This normalized residual is equivalent to the Mahalanobis distance squared.
    For residuals with no covariance (all non-diagonal covariance elements are zero) this
    is exactly equivalent to chi2.

    If the off-diagonal covariance elements (the covariate terms) are missing (represented by NaNs),
    then they will assumed to be 0.0.

    Parameters
    ----------
    residuals : `~numpy.ndarray` (N, D)
        Array of N residuals in D dimensions (observed - predicted).
    covariances : `~numpy.ndarray` (N, D, D)
        Array of N covariance matrices in D dimensions.

    Returns
    -------
    chi2 : `~numpy.ndarray` (N)
        Array of chi2 values for each residual and covariance pair.

    References
    ----------
    [1] Carpino, M. et al. (2003). Error statistics of asteroid optical astrometric observations.
        Icarus, 166, 248-270. https://doi.org/10.1016/S0019-1035(03)00051-4
    """
    # Raise error if any of the diagonal elements are nan
    if np.any(np.isnan(np.diagonal(covariances, axis1=1, axis2=2))):
        raise ValueError("Covariance matrix has NaNs on the diagonal.")

    # Warn if any of the non-diagonal elements are nan
    if np.any(np.isnan(covariances)):
        warnings.warn(
            "Covariance matrix has NaNs on the off-diagonal (these will be assumed to be 0.0).",
            UserWarning,
        )
        covariances = np.where(np.isnan(covariances), 0.0, covariances)

    W = np.linalg.inv(covariances)
    chi2 = np.einsum("ij,ji->i", np.einsum("ij,ijk->ik", residuals, W), residuals.T)
    return chi2


def calculate_reduced_chi2(residuals: Residuals, parameters: int) -> float:
    """
    Calculate the reduced chi2 for a set of residuals.

    Parameters
    ----------
    residuals : `~adam_core.coordinates.residuals.Residuals`
        Residuals.
    parameters : int
        Number of parameters in the model.

    Returns
    -------
    reduced_chi2 : float
        Reduced chi2.
    """
    chi2_total = residuals.chi2.to_numpy().sum()
    dof_total = residuals.dof.to_numpy().sum() - parameters
    return chi2_total / dof_total


def _batch_coords_and_covariances(
    coords: npt.NDArray[np.float64], covariances: npt.NDArray[np.float64]
) -> Tuple[
    List[npt.NDArray[np.float64]],
    List[npt.NDArray[np.float64]],
    List[npt.NDArray[np.float64]],
    List[npt.NDArray[np.float64]],
]:
    """
    Batch coordinates and covariances into groups that have the same
    number of dimensions that have missing values (represented by NaNs).

    Parameters
    ----------
    coords : `~numpy.ndarray` (N, D)
        Array of N coordinates in D dimensions.
    covariances : `~numpy.ndarray` (N, D, D)
        Array of N covariance matrices in D dimensions.

    Returns
    -------
    batch_indices : List[`~numpy.ndarray` (<=N)]
        List of arrays of indices into the original arrays that correspond to
        coordinates with the same number of dimensions that have missing values.
    batch_dimensions : List[`~numpy.ndarray` (<=D)]
        List of arrays of dimensions that have values for each batch.
    batch_coords : List[`~numpy.ndarray` (<=N, <=D)]
        Array of N coordinates in D dimensions with missing dimensions removed.
    batch_covariances : List[`~numpy.ndarray` (<=N, <=D, <=D)]
        Array of N covariance matrices in D dimensions with missing dimensions removed.
    """
    N, D = coords.shape
    assert covariances.shape == (N, D, D)

    # Find the indices of the coordinates that have the same missing dimensions
    # (if any) and select the coordinates and covariances for those indices
    nans = np.isnan(coords)
    batch_masks = np.unique(nans, axis=0)
    indices = np.arange(0, N, dtype=np.int64)

    batch_indices: List[np.ndarray] = []
    batch_dimensions: List[np.ndarray] = []
    batch_coords: List[np.ndarray] = []
    batch_covariances: List[np.ndarray] = []
    for batch in batch_masks:
        # Find the indices of the coordinates that have the same missing dimensions
        # and select the coordinates and covariances for those indices
        batch_mask_i = indices[np.where(np.all(nans == batch, axis=1))[0]]
        coords_i = coords[batch_mask_i]
        covariances_i = covariances[batch_mask_i]

        # Remove the missing dimensions from the coordinates and covariances
        dimensions_with_values = np.where(~batch)[0]
        coords_i = coords_i[:, dimensions_with_values]
        covariances_i = covariances_i[:, dimensions_with_values, :]
        covariances_i = covariances_i[:, :, dimensions_with_values]

        batch_indices.append(batch_mask_i)
        batch_dimensions.append(dimensions_with_values)
        batch_coords.append(coords_i)
        batch_covariances.append(covariances_i)

    return batch_indices, batch_dimensions, batch_coords, batch_covariances



import numpy as np
import spiceypy as sp
from astropy import units as u

from ...coordinates.cartesian import CartesianCoordinates
from ...coordinates.origin import Origin
from ...orbits import Orbits
from ...time import Timestamp
from ..propagation import _propagate_2body, _propagate_2body_vmap, propagate_2body


def test__propagate_2body_against_spice_elliptical(orbital_elements):
    # Test propagation against SPICE for elliptical orbits.

    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(cartesian_elements))])
    mu = origin.mu()

    dts = np.arange(0, 10000, 10)
    for i, (t0_i, cartesian_i) in enumerate(zip(t0, cartesian_elements)):

        spice_propagated = np.empty((len(dts), 6))
        cartesian_propagated = np.empty((len(dts), 6))

        for j, dt_i in enumerate(dts):

            spice_propagated[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )
            cartesian_propagated[j] = _propagate_2body(
                cartesian_i, t0_i, t0_i + dt_i, mu[i]
            )

        # Calculate difference between SPICE and adam_core
        diff = cartesian_propagated - spice_propagated

        # Calculate offset in position in cm
        r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
        # Calculate offset in velocity in mm/s
        v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

        # Assert positions are to within 10 cm
        np.testing.assert_array_less(r_diff, 10)
        # Assert velocities are to within 1 mm/s
        np.testing.assert_array_less(v_diff, 1)


def test__propagate_2body_against_spice_hyperbolic(orbital_elements):
    # Test propagation against SPICE for hyperbolic orbits.

    orbital_elements = orbital_elements[orbital_elements["e"] > 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(cartesian_elements))])
    mu = origin.mu()

    dts = np.arange(0, 10000, 10)
    for i, (t0_i, cartesian_i) in enumerate(zip(t0, cartesian_elements)):

        spice_propagated = np.empty((len(dts), 6))
        cartesian_propagated = np.empty((len(dts), 6))

        for j, dt_i in enumerate(dts):

            spice_propagated[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )
            cartesian_propagated[j] = _propagate_2body(
                cartesian_i, t0_i, t0_i + dt_i, mu[i]
            )

        # Calculate difference between SPICE and adam_core
        diff = cartesian_propagated - spice_propagated

        # Calculate offset in position in cm
        r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
        # Calculate offset in velocity in mm/s
        v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

        # Assert positions are to within 10 cm
        np.testing.assert_array_less(r_diff, 10)
        # Assert velocities are to within 1 mm/s
        np.testing.assert_array_less(v_diff, 1)


def test_propagate_2body_against_spice_elliptical(orbital_elements):
    # Test propagation against SPICE for elliptical orbits.

    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(cartesian_elements))])
    mu = origin.mu()

    # Create orbits object
    orbits = Orbits.from_kwargs(
        orbit_id=orbital_elements["targetname"].values,
        object_id=orbital_elements["targetname"].values,
        coordinates=CartesianCoordinates.from_kwargs(
            x=cartesian_elements[:, 0],
            y=cartesian_elements[:, 1],
            z=cartesian_elements[:, 2],
            vx=cartesian_elements[:, 3],
            vy=cartesian_elements[:, 4],
            vz=cartesian_elements[:, 5],
            time=Timestamp.from_mjd(
                t0,
                scale="tdb",
            ),
            origin=origin,
            frame="ecliptic",
        ),
    )

    # Set propagation times (same for all orbits)
    times = Timestamp.from_mjd(
        t0.min() + np.arange(0, 10000, 10),
        scale="tdb",
    )

    # Propagate orbits with SPICE and accumulate results
    spice_propagated = []
    times_mjd = times.mjd()
    for i, cartesian_i in enumerate(cartesian_elements):

        # Calculate dts from t0 for this orbit (each orbit's t0 is different)
        # but the final times we are propagating to are the same for all orbits
        dts = times_mjd - t0[i]
        spice_propagated_i = np.empty((len(dts), 6))
        for j, dt_i in enumerate(dts):
            spice_propagated_i[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )

        spice_propagated.append(spice_propagated_i)

    spice_propagated = np.vstack(spice_propagated)

    # Propagate orbits with adam_core
    orbits_propagated = propagate_2body(orbits, times)

    # Calculate difference between SPICE and adam_core
    diff = orbits_propagated.coordinates.values - spice_propagated

    # Calculate offset in position in cm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 10 cm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 1 mm/s
    np.testing.assert_array_less(v_diff, 1)


def test_propagate_2body_against_spice_hyperbolic(orbital_elements):
    # Test propagation against SPICE for hyperbolic orbits.
    orbital_elements = orbital_elements[orbital_elements["e"] > 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(cartesian_elements))])
    mu = origin.mu()

    # Create orbits object
    orbits = Orbits.from_kwargs(
        orbit_id=orbital_elements["targetname"].values,
        object_id=orbital_elements["targetname"].values,
        coordinates=CartesianCoordinates.from_kwargs(
            x=cartesian_elements[:, 0],
            y=cartesian_elements[:, 1],
            z=cartesian_elements[:, 2],
            vx=cartesian_elements[:, 3],
            vy=cartesian_elements[:, 4],
            vz=cartesian_elements[:, 5],
            time=Timestamp.from_mjd(
                t0,
                scale="tdb",
            ),
            origin=origin,
            frame="ecliptic",
        ),
    )

    # Set propagation times (same for all orbits)
    times = Timestamp.from_mjd(
        t0.min() + np.arange(0, 10000, 10),
        scale="tdb",
    )

    # Propagate orbits with SPICE and accumulate results
    spice_propagated = []
    times_mjd = times.mjd()
    for i, cartesian_i in enumerate(cartesian_elements):

        # Calculate dts from t0 for this orbit (each orbit's t0 is different)
        # but the final times we are propagating to are the same for all orbits
        dts = times_mjd - t0[i]
        spice_propagated_i = np.empty((len(dts), 6))
        for j, dt_i in enumerate(dts):
            spice_propagated_i[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )

        spice_propagated.append(spice_propagated_i)

    spice_propagated = np.vstack(spice_propagated)

    # Propagate orbits with adam_core
    orbits_propagated = propagate_2body(orbits, times)

    # Calculate difference between SPICE and adam_core
    diff = orbits_propagated.coordinates.values - spice_propagated

    # Calculate offset in position in cm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 10 cm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 1 mm/s
    np.testing.assert_array_less(v_diff, 1)


def test_propagate_2body_against_spice_elliptical_barycentric(
    orbital_elements_barycentric,
):
    # Test propagation against SPICE for elliptical orbits.
    orbital_elements = orbital_elements_barycentric[
        orbital_elements_barycentric["e"] < 1.0
    ]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(
        code=["SOLAR_SYSTEM_BARYCENTER" for i in range(len(cartesian_elements))]
    )
    mu = origin.mu()

    # Create orbits object
    orbits = Orbits.from_kwargs(
        orbit_id=orbital_elements["targetname"].values,
        object_id=orbital_elements["targetname"].values,
        coordinates=CartesianCoordinates.from_kwargs(
            x=cartesian_elements[:, 0],
            y=cartesian_elements[:, 1],
            z=cartesian_elements[:, 2],
            vx=cartesian_elements[:, 3],
            vy=cartesian_elements[:, 4],
            vz=cartesian_elements[:, 5],
            time=Timestamp.from_mjd(
                t0,
                scale="tdb",
            ),
            origin=origin,
            frame="ecliptic",
        ),
    )

    # Set propagation times (same for all orbits)
    times = Timestamp.from_mjd(
        t0.min() + np.arange(0, 10000, 10),
        scale="tdb",
    )

    # Propagate orbits with SPICE and accumulate results
    spice_propagated = []
    times_mjd = times.mjd()
    for i, cartesian_i in enumerate(cartesian_elements):

        # Calculate dts from t0 for this orbit (each orbit's t0 is different)
        # but the final times we are propagating to are the same for all orbits
        dts = times_mjd - t0[i]
        spice_propagated_i = np.empty((len(dts), 6))
        for j, dt_i in enumerate(dts):
            spice_propagated_i[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )

        spice_propagated.append(spice_propagated_i)

    spice_propagated = np.vstack(spice_propagated)

    # Propagate orbits with adam_core
    orbits_propagated = propagate_2body(orbits, times)

    # Calculate difference between SPICE and adam_core
    diff = orbits_propagated.coordinates.values - spice_propagated

    # Calculate offset in position in cm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 10 cm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 1 mm/s
    np.testing.assert_array_less(v_diff, 1)


def test_propagate_2body_against_spice_hyperbolic_barycentric(
    orbital_elements_barycentric,
):
    # Test propagation against SPICE for hyperbolic orbits.

    orbital_elements = orbital_elements_barycentric[
        orbital_elements_barycentric["e"] > 1.0
    ]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(
        code=["SOLAR_SYSTEM_BARYCENTER" for i in range(len(cartesian_elements))]
    )
    mu = origin.mu()

    # Create orbits object
    orbits = Orbits.from_kwargs(
        orbit_id=orbital_elements["targetname"].values,
        object_id=orbital_elements["targetname"].values,
        coordinates=CartesianCoordinates.from_kwargs(
            x=cartesian_elements[:, 0],
            y=cartesian_elements[:, 1],
            z=cartesian_elements[:, 2],
            vx=cartesian_elements[:, 3],
            vy=cartesian_elements[:, 4],
            vz=cartesian_elements[:, 5],
            time=Timestamp.from_mjd(
                t0,
                scale="tdb",
            ),
            origin=origin,
            frame="ecliptic",
        ),
    )

    # Set propagation times (same for all orbits)
    times = Timestamp.from_mjd(
        t0.min() + np.arange(0, 10000, 10),
        scale="tdb",
    )

    # Propagate orbits with SPICE and accumulate results
    spice_propagated = []
    times_mjd = times.mjd()
    for i, cartesian_i in enumerate(cartesian_elements):

        # Calculate dts from t0 for this orbit (each orbit's t0 is different)
        # but the final times we are propagating to are the same for all orbits
        dts = times_mjd - t0[i]
        spice_propagated_i = np.empty((len(dts), 6))
        for j, dt_i in enumerate(dts):
            spice_propagated_i[j] = sp.prop2b(
                mu[i], np.ascontiguousarray(cartesian_i), dt_i
            )

        spice_propagated.append(spice_propagated_i)

    spice_propagated = np.vstack(spice_propagated)

    # Propagate orbits with adam_core
    orbits_propagated = propagate_2body(orbits, times)

    # Calculate difference between SPICE and adam_core
    diff = orbits_propagated.coordinates.values - spice_propagated

    # Calculate offset in position in cm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.cm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 10 cm
    np.testing.assert_array_less(r_diff, 10)
    # Assert velocities are to within 1 mm/s
    np.testing.assert_array_less(v_diff, 1)


def test_benchmark__propagate_2body(benchmark, orbital_elements):
    # Benchmark _propagate_2body with a single orbit propagated forward 1 day
    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))])
    mu = origin.mu()

    benchmark(_propagate_2body, cartesian_elements[0], t0[0], t0[0] + 1, mu[0])


def test_benchmark__propagate_2body_vmap(benchmark, orbital_elements):
    # Benchmark the vectorized map version of _propagate_2body with a single
    # orbit propagated forward 1 day
    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))])
    mu = origin.mu()

    benchmark(
        _propagate_2body_vmap,
        cartesian_elements[:1],
        t0[:1],
        t0[:1] + 1,
        mu[:1],
        1000,
        1e-14,
    )


def test_benchmark_propagate_2body(benchmark, orbital_elements):
    # Benchmark propagate_2body with a single orbit propagated forward 1 day
    # This function appears to add substantial overhead, so we benchmark it
    # separately from _propagate_2body
    orbital_elements = orbital_elements[orbital_elements["e"] < 1.0]
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    t0 = orbital_elements["mjd_tdb"].values
    # Create orbits object
    orbits = Orbits.from_kwargs(
        orbit_id=orbital_elements["targetname"].values,
        object_id=orbital_elements["targetname"].values,
        coordinates=CartesianCoordinates.from_kwargs(
            x=cartesian_elements[:, 0],
            y=cartesian_elements[:, 1],
            z=cartesian_elements[:, 2],
            vx=cartesian_elements[:, 3],
            vy=cartesian_elements[:, 4],
            vz=cartesian_elements[:, 5],
            time=Timestamp.from_mjd(
                t0,
                scale="tdb",
            ),
            origin=Origin.from_kwargs(
                code=["SUN" for i in range(len(cartesian_elements))]
            ),
            frame="ecliptic",
        ),
    )
    times = Timestamp.from_mjd(
        [t0.min() + 1],
        scale="tdb",
    )
    benchmark(propagate_2body, orbits[0], times=times)



from ..associations import Associations
from ..detections import PointSourceDetections


def test_associations_link_to_detections():
    # Test that we can link associations to detections
    detections = PointSourceDetections.from_kwargs(
        id=["d1", "d2", "d3", "d4", "d5"],
        exposure_id=["e1", "e1", "e2", "e1", "e2"],
        time=None,
        ra=[0, 1, 2, 3, 4],
        dec=[5, 6, 7, 8, 9],
        mag=[10, 11, 12, 13, 14],
    )

    associations = Associations.from_kwargs(
        detection_id=["d1", "d2", "d3", "d4", "d5"],
        object_id=["o1", "o1", "o2", "o2", None],
    )

    link = associations.link_to_detections(detections)
    assert link.left_table == associations
    assert link.right_table == detections

    for i in range(5):
        associations_i, detetections_i = link.select(f"d{i + 1}")
        assert len(associations_i) == 1
        assert len(detetections_i) == 1


def test_associations_group_by_object():
    # Test that we can group associations by object
    associations = Associations.from_kwargs(
        detection_id=["d1", "d2", "d3", "d4", "d5"],
        object_id=["o1", "o1", "o2", "o2", None],
    )

    groups = list(associations.group_by_object())
    assert len(groups) == 3
    assert groups[0].object_id.to_pylist() == ["o1", "o1"]
    assert groups[0].detection_id.to_pylist() == ["d1", "d2"]

    assert groups[1].object_id.to_pylist() == ["o2", "o2"]
    assert groups[1].detection_id.to_pylist() == ["d3", "d4"]

    # What is the expected return here?
    assert groups[2].object_id.to_pylist() == [None]



import numpy as np
import pytest

from ...coordinates.origin import OriginCodes
from ...observers import get_observer_state
from ...time import Timestamp


@pytest.mark.parametrize(
    "times",
    [1, 10000],
    ids=lambda x: f"times={x},",
)
@pytest.mark.parametrize("code", ["X05", "500"], ids=lambda x: f"code={x},")
@pytest.mark.parametrize(
    "frame", ["equatorial", "ecliptic"], ids=lambda x: f"frame={x},"
)
@pytest.mark.parametrize(
    "origin",
    [OriginCodes.SUN, OriginCodes.SOLAR_SYSTEM_BARYCENTER],
    ids=lambda x: f"origin={x.name},",
)
@pytest.mark.benchmark(group="observer_states")
def test_benchmark_get_observer_state(benchmark, times, code, frame, origin):
    # We can expect needing to get the observer states for duplicated observation
    # times in the future, so we should benchmark this case
    if times == 10000:
        # 1000 times duplicated 10 times
        repeats = 10
    else:
        # 1 time not duplicated
        repeats = 1

    times_array = np.concatenate(
        [np.arange(60000, 60000 + np.minimum(times, 1000), 1) for i in range(repeats)]
    )
    times = Timestamp.from_mjd(np.sort(times_array), scale="tdb")

    result = benchmark(
        get_observer_state,
        code,
        times,
        frame=frame,
        origin=origin,
    )
    assert len(result) == len(times)



import numpy as np
import pytest

from ...coordinates.origin import Origin
from ...coordinates.residuals import Residuals
from ...coordinates.spherical import SphericalCoordinates
from ...observers.observers import Observers
from ...time.time import Timestamp
from ..differential_correction import OrbitDeterminationObservations
from ..fitted_orbits import FittedOrbitMembers
from ..outliers import remove_lowest_probability_observation


def test_remove_lowest_probability_observation():
    # Test that remove lowest probability observation correctly identifies and
    # removes the worst outlier

    # Create orbit members with 4 observations, one of which is a bad outlier
    orbit_members = FittedOrbitMembers.from_kwargs(
        orbit_id=["orbit01", "orbit01", "orbit01", "orbit01"],
        obs_id=["obs01", "obs02", "obs03", "obs04"],
        residuals=Residuals.from_kwargs(
            values=[
                [np.nan, 0.1, 0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.1, 0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.1, 0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.1, 0.1, np.nan, np.nan, np.nan],
            ],
            dof=np.full(4, 2),
            probability=[0.2, 0.1, 0.0001, 0.3],
        ),
    )

    # Create observations with 4 observations
    time = Timestamp.from_mjd(np.arange(59000, 59004), scale="utc")
    observations = OrbitDeterminationObservations.from_kwargs(
        id=["obs01", "obs02", "obs03", "obs04"],
        coordinates=SphericalCoordinates.from_kwargs(
            lon=np.random.rand(4),
            lat=np.random.rand(4),
            origin=Origin.from_kwargs(code=np.full(4, "500", dtype="object")),
            time=time,
        ),
        observers=Observers.from_code("500", time),
    )

    obs_id, observations = remove_lowest_probability_observation(
        orbit_members, observations
    )
    assert len(observations) == 3
    assert observations.id.tolist() == ["obs01", "obs02", "obs04"]
    assert obs_id == "obs03"


def test_remove_lowest_probability_observation_multiple():
    # Test that remove lowest probability observation correctly identifies and
    # removes the worst outlier when there are multiple observations with the same
    # probability

    # Create orbit members with 4 observations, one of which is a bad outlier
    orbit_members = FittedOrbitMembers.from_kwargs(
        orbit_id=["orbit01", "orbit01", "orbit01", "orbit01"],
        obs_id=["obs01", "obs02", "obs03", "obs04"],
        residuals=Residuals.from_kwargs(
            values=[
                [np.nan, 0.1, -0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.6, -0.6, np.nan, np.nan, np.nan],
                [np.nan, 0.3, -0.3, np.nan, np.nan, np.nan],
                [np.nan, 0.5, -0.5, np.nan, np.nan, np.nan],
            ],
            dof=np.full(4, 2),
            probability=[0.2, 0.2, 0.2, 0.2],
        ),
    )

    # Create observations with 4 observations
    time = Timestamp.from_mjd(np.arange(59000, 59004), scale="utc")
    observations = OrbitDeterminationObservations.from_kwargs(
        id=["obs01", "obs02", "obs03", "obs04"],
        coordinates=SphericalCoordinates.from_kwargs(
            lon=np.random.rand(4),
            lat=np.random.rand(4),
            origin=Origin.from_kwargs(code=np.full(4, "500", dtype="object")),
            time=time,
        ),
        observers=Observers.from_code("500", time),
    )

    obs_id, observations = remove_lowest_probability_observation(
        orbit_members, observations
    )
    assert len(observations) == 3
    assert observations.id.tolist() == ["obs01", "obs03", "obs04"]
    assert obs_id == "obs02"


def test_remove_lowest_probability_observation_assertions():

    # Create observations with 4 observations
    time = Timestamp.from_mjd(np.arange(59000, 59004), scale="utc")
    observations = OrbitDeterminationObservations.from_kwargs(
        id=["obs01", "obs02", "obs03", "obs04"],
        coordinates=SphericalCoordinates.from_kwargs(
            lon=np.random.rand(4),
            lat=np.random.rand(4),
            origin=Origin.from_kwargs(code=np.full(4, "500", dtype="object")),
            time=time,
        ),
        observers=Observers.from_code("500", time),
    )

    # Create orbit members with two orbits
    orbit_members = FittedOrbitMembers.from_kwargs(
        orbit_id=["orbit01", "orbit01", "orbit01", "orbit02"],
        obs_id=["obs01", "obs02", "obs03", "obs04"],
        residuals=Residuals.from_kwargs(
            values=[
                [np.nan, 0.1, -0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.6, -0.6, np.nan, np.nan, np.nan],
                [np.nan, 0.3, -0.3, np.nan, np.nan, np.nan],
                [np.nan, 0.5, -0.5, np.nan, np.nan, np.nan],
            ],
            dof=np.full(4, 2),
            probability=[0.2, 0.2, 0.2, 0.2],
        ),
    )

    with pytest.raises(
        AssertionError, match=r"Orbit members must only contain one orbit"
    ):
        obs_id, observations = remove_lowest_probability_observation(
            orbit_members, observations
        )

    # Create orbit members with 4 observations, two of which do not
    # exit in the observations
    orbit_members = FittedOrbitMembers.from_kwargs(
        orbit_id=["orbit01", "orbit01", "orbit01", "orbit01"],
        obs_id=["obs03", "obs04", "obs05", "obs06"],
        residuals=Residuals.from_kwargs(
            values=[
                [np.nan, 0.1, -0.1, np.nan, np.nan, np.nan],
                [np.nan, 0.6, -0.6, np.nan, np.nan, np.nan],
                [np.nan, 0.3, -0.3, np.nan, np.nan, np.nan],
                [np.nan, 0.5, -0.5, np.nan, np.nan, np.nan],
            ],
            dof=np.full(4, 2),
            probability=[0.2, 0.2, 0.2, 0.2],
        ),
    )

    with pytest.raises(
        AssertionError, match=r"Observations must contain all orbit member observations"
    ):
        obs_id, observations = remove_lowest_probability_observation(
            orbit_members, observations
        )



from __future__ import annotations

# flake8: noqa: F401
from .cartesian import CartesianCoordinates
from .cometary import CometaryCoordinates
from .covariances import CoordinateCovariances
from .keplerian import KeplerianCoordinates
from .origin import Origin, OriginCodes
from .residuals import Residuals
from .spherical import SphericalCoordinates
from .transform import transform_coordinates



import numpy as np
import pytest
import quivr as qv

from ..utils import _assert_times_almost_equal, _iterate_chunks


class SampleTable(qv.Table):
    a = qv.Float64Column()


def test__iterate_chunks():
    # Test that _iterate_chunks works with numpy arrays and lists
    a = np.arange(0, 10)
    for chunk in _iterate_chunks(a, 2):
        assert len(chunk) == 2

    a = [i for i in range(10)]
    for chunk in _iterate_chunks(a, 2):
        assert len(chunk) == 2


def test__iterate_chunks_table():
    # Test that _iterate_chunks works with quivr tables
    table = SampleTable.from_kwargs(a=np.arange(0, 11))
    chunks = list(_iterate_chunks(table, 2))
    assert len(chunks) == 6

    for i, chunk in enumerate(chunks):
        if i != 5:
            assert len(chunk) == 2
            assert isinstance(chunk, SampleTable)
            np.testing.assert_equal(chunk.a.to_numpy(), np.arange(i * 2, i * 2 + 2))
        else:
            assert len(chunk) == 1
            assert isinstance(chunk, SampleTable)
            np.testing.assert_equal(chunk.a.to_numpy(), np.arange(i * 2, i * 2 + 1))


def test__assert_times_almost_equal():
    have = np.array([1.0, 2.0, 3.0])
    want = np.array([1.0, 2.0, 3.0])

    _assert_times_almost_equal(have, want, tolerance=1.0)

    with pytest.raises(ValueError):
        have = np.array([1.0, 2.0, 3.0])
        want = np.array([1.0, 2.0, 3.0])

        # Offset have by 2 ms
        have += 2 / 86800 / 1000
        _assert_times_almost_equal(have, want, tolerance=1.0)



from __future__ import annotations

from typing import Iterator, Literal

import pyarrow as pa
import pyarrow.compute as pc
import quivr as qv
from quivr.validators import and_, ge, le

from ..coordinates.origin import OriginCodes
from ..observers import observers, state
from ..time import Timestamp


class Exposures(qv.Table):
    """
    Exposures is a table of data about exposures that provide point source observations.
    """

    id = qv.LargeStringColumn()
    start_time = Timestamp.as_column()
    duration = qv.Float64Column(validator=and_(ge(0), le(3600)))
    filter = qv.DictionaryColumn(index_type=pa.int32(), value_type=pa.string())

    observatory_code = qv.LargeStringColumn()

    def group_by_observatory_code(self) -> Iterator[tuple[str, Exposures]]:
        """
        Groups the exposures by observatory code.
        """
        unique_codes = self.observatory_code.unique()
        for code in unique_codes:
            mask = pc.equal(self.observatory_code, code)
            yield code, self.apply_mask(mask)

    def observers(
        self,
        frame: Literal["ecliptic", "equatorial"] = "ecliptic",
        origin: OriginCodes = OriginCodes.SUN,
    ) -> observers.Observers:
        """
        Return the observer location at each exposure midpoint.
        """

        # bunch of bookkeeping here to return states with the same
        # indexing as self
        coords = []
        indices = []
        codes = []

        unique_codes = self.observatory_code.unique().to_pylist()
        for code in unique_codes:
            mask = pc.equal(self.observatory_code, code)
            exposures_for_code = self.apply_mask(mask)
            indices_for_code = pc.indices_nonzero(mask)

            times_for_code = exposures_for_code.midpoint()
            coords_for_code = state.get_observer_state(
                code, times_for_code, frame=frame, origin=origin
            )

            coords.append(coords_for_code)
            indices.append(indices_for_code)
            codes.append(pa.array([code] * len(coords_for_code)))

        observers_table = observers.Observers.from_kwargs(
            code=pa.concat_arrays(codes),
            coordinates=qv.concatenate(coords),
        )

        indices = pa.concat_arrays(indices)
        original_ordering = pc.array_sort_indices(indices)
        return observers_table.take(original_ordering)

    def midpoint(self) -> Timestamp:
        """
        Returns the midpoint of the exposure.
        """
        return self.start_time.add_seconds(pc.divide(self.duration, 2.0))



from ...time import Timestamp
from ..cartesian import CartesianCoordinates
from ..origin import Origin
from ..transform import cartesian_to_frame, transform_coordinates
from .test_transforms_translation import assert_coords_equal


def test_cartesian_to_frame(orbital_elements, orbital_elements_equatorial):
    # Test cartesian_to_frame correctly converts between ecliptic and
    # equatorial cartesian coordinates

    cartesian_coordinates_ecliptic = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(orbital_elements["mjd_tdb"].values, scale="tdb"),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_equatorial = CartesianCoordinates.from_kwargs(
        x=orbital_elements_equatorial["x"].values,
        y=orbital_elements_equatorial["y"].values,
        z=orbital_elements_equatorial["z"].values,
        vx=orbital_elements_equatorial["vx"].values,
        vy=orbital_elements_equatorial["vy"].values,
        vz=orbital_elements_equatorial["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_equatorial["mjd_tdb"].values, scale="tdb"
        ),
        frame="equatorial",
        origin=Origin.from_kwargs(
            code=["SUN" for i in range(len(orbital_elements_equatorial))]
        ),
    )

    # Convert equatorial coordinates to ecliptic
    cartesian_coordinates_ecliptic_actual = cartesian_to_frame(
        cartesian_coordinates_equatorial,
        "ecliptic",
    )
    assert cartesian_coordinates_ecliptic_actual.frame == "ecliptic"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_ecliptic_actual, cartesian_coordinates_ecliptic
    )

    # Convert ecliptic coordinates to equatorial
    cartesian_coordinates_equatorial_actual = cartesian_to_frame(
        cartesian_coordinates_ecliptic,
        "equatorial",
    )
    assert cartesian_coordinates_equatorial_actual.frame == "equatorial"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_equatorial_actual, cartesian_coordinates_equatorial
    )


def test_transform_coordinates_frame(orbital_elements, orbital_elements_equatorial):
    # Test transform_coordinates correctly converts between ecliptic and
    # equatorial cartesian coordinates

    cartesian_coordinates_ecliptic = CartesianCoordinates.from_kwargs(
        x=orbital_elements["x"].values,
        y=orbital_elements["y"].values,
        z=orbital_elements["z"].values,
        vx=orbital_elements["vx"].values,
        vy=orbital_elements["vy"].values,
        vz=orbital_elements["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN" for i in range(len(orbital_elements))]),
    )

    cartesian_coordinates_equatorial = CartesianCoordinates.from_kwargs(
        x=orbital_elements_equatorial["x"].values,
        y=orbital_elements_equatorial["y"].values,
        z=orbital_elements_equatorial["z"].values,
        vx=orbital_elements_equatorial["vx"].values,
        vy=orbital_elements_equatorial["vy"].values,
        vz=orbital_elements_equatorial["vz"].values,
        time=Timestamp.from_mjd(
            orbital_elements_equatorial["mjd_tdb"].values,
            scale="tdb",
        ),
        frame="equatorial",
        origin=Origin.from_kwargs(
            code=["SUN" for i in range(len(orbital_elements_equatorial))]
        ),
    )

    # Convert equatorial coordinates to ecliptic
    cartesian_coordinates_ecliptic_actual = transform_coordinates(
        cartesian_coordinates_equatorial,
        frame_out="ecliptic",
    )
    assert cartesian_coordinates_ecliptic_actual.frame == "ecliptic"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_ecliptic_actual, cartesian_coordinates_ecliptic
    )

    # Convert ecliptic coordinates to equatorial
    cartesian_coordinates_equatorial_actual = transform_coordinates(
        cartesian_coordinates_ecliptic,
        frame_out="equatorial",
    )
    assert cartesian_coordinates_equatorial_actual.frame == "equatorial"

    # Test that the two coordinates are equal to 10 mm in position and 10 nm/s in velocity
    assert_coords_equal(
        cartesian_coordinates_equatorial_actual, cartesian_coordinates_equatorial
    )



import numpy as np

from ...coordinates.keplerian import KeplerianCoordinates
from ...coordinates.origin import Origin
from ...orbits import Orbits
from ...time import Timestamp
from ..moid import calculate_moid


def test_calculate_moid_circular_orbits():
    # Test for two coplanar circular orbits that their MOID
    # is equal to the difference of their semi-major axes
    circular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[1.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    circular_3 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[3.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(circular_1, circular_3)
    np.testing.assert_allclose(moid_13, 2.0, rtol=0, atol=1e-12)

    moid_31, moid_time_31 = calculate_moid(circular_3, circular_1)
    np.testing.assert_allclose(moid_31, 2.0, rtol=0, atol=1e-12)

    np.testing.assert_allclose(moid_13, moid_31, rtol=0, atol=1e-12)

    # Test that the MOID of two orbits with relative inclination of 90 degrees
    # is equal to the difference of their semi-major axes
    circular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[1.0],
            e=[0.0],
            i=[0.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    inclined_3 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=KeplerianCoordinates.from_kwargs(
            a=[3.0],
            e=[0.0],
            i=[90.0],
            ap=[0.0],
            raan=[0.0],
            M=[0.0],
            time=Timestamp.from_mjd([59000.0], scale="tdb"),
            frame="ecliptic",
            origin=Origin.from_kwargs(code=["SUN"]),
        ).to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(circular_1, inclined_3)
    np.testing.assert_allclose(moid_13, 2.0, rtol=0, atol=1e-12)

    moid_31, moid_time_31 = calculate_moid(inclined_3, circular_1)
    np.testing.assert_allclose(moid_31, 2.0, rtol=0, atol=1e-12)

    np.testing.assert_allclose(moid_13, moid_31, rtol=0, atol=1e-12)


def test_calculate_moid_noncircular_orbits():

    # Test that the MOID of two coplanar but non-circular orbits located on
    # the x-axis at same time is equal to the difference
    # of their periapse distances
    j2000 = Timestamp.from_jd([2451545.0], scale="tt")

    noncircular_keplerian_1 = KeplerianCoordinates.from_kwargs(
        a=[2.0],
        e=[0.2],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=noncircular_keplerian_1.to_cartesian(),
    )

    noncircular_keplerian_2 = KeplerianCoordinates.from_kwargs(
        a=[5.0],
        e=[0.2],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_2 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=noncircular_keplerian_2.to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(noncircular_1, noncircular_2)
    np.testing.assert_allclose(
        moid_13,
        noncircular_keplerian_2.q[0] - noncircular_keplerian_1.q[0],
        rtol=0,
        atol=1e-12,
    )

    moid_31, moid_time_31 = calculate_moid(noncircular_2, noncircular_1)
    np.testing.assert_allclose(
        moid_31,
        noncircular_keplerian_2.q[0] - noncircular_keplerian_1.q[0],
        rtol=0,
        atol=1e-12,
    )

    # Test again with two orbits that overlap at some point
    noncircular_keplerian_1 = KeplerianCoordinates.from_kwargs(
        a=[2.0],
        e=[0.1],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_1 = Orbits.from_kwargs(
        orbit_id=["a"],
        coordinates=noncircular_keplerian_1.to_cartesian(),
    )

    noncircular_keplerian_2 = KeplerianCoordinates.from_kwargs(
        a=[3.0],
        e=[0.9],
        i=[0.0],
        ap=[0.0],
        raan=[0.0],
        M=[0.0],
        time=j2000,
        frame="ecliptic",
        origin=Origin.from_kwargs(code=["SUN"]),
    )
    noncircular_2 = Orbits.from_kwargs(
        orbit_id=["b"],
        coordinates=noncircular_keplerian_2.to_cartesian(),
    )

    moid_13, moid_time_13 = calculate_moid(noncircular_1, noncircular_2)
    np.testing.assert_allclose(moid_13, 0.0, rtol=0, atol=1e-6)

    moid_31, moid_time_31 = calculate_moid(noncircular_2, noncircular_1)
    np.testing.assert_allclose(moid_31, 0.0, rtol=0, atol=1e-6)



from __future__ import annotations

import logging
from typing import Literal, Optional, Union

import jax.numpy as jnp
import numpy as np
import pyarrow.compute as pc
from jax import config, jit, lax, vmap

from ..constants import Constants as c
from . import types
from .cartesian import CartesianCoordinates
from .cometary import CometaryCoordinates
from .keplerian import KeplerianCoordinates
from .origin import OriginCodes
from .spherical import SphericalCoordinates

config.update("jax_enable_x64", True)
config.update("jax_platform_name", "cpu")

TRANSFORM_EQ2EC = np.zeros((6, 6))
TRANSFORM_EQ2EC[0:3, 0:3] = c.TRANSFORM_EQ2EC
TRANSFORM_EQ2EC[3:6, 3:6] = c.TRANSFORM_EQ2EC
TRANSFORM_EC2EQ = TRANSFORM_EQ2EC.T

logger = logging.getLogger(__name__)


CoordinatesClasses = (
    CartesianCoordinates,
    KeplerianCoordinates,
    CometaryCoordinates,
    SphericalCoordinates,
)


Z_AXIS = jnp.array([0.0, 0.0, 1.0])
FLOAT_TOLERANCE = 1e-15


@jit
def _cartesian_to_spherical(
    coords_cartesian: Union[np.ndarray, jnp.ndarray]
) -> jnp.ndarray:
    """
    Convert a single Cartesian coordinate to a spherical coordinate.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of distance.
        y : y-position in units of distance.
        z : z-position in units of distance.
        vx : x-velocity in the same units of x per arbitrary unit of time.
        vy : y-velocity in the same units of y per arbitrary unit of time.
        vz : z-velocity in the same units of z per arbitrary unit of time.

    Returns
    -------
    coords_spherical : `~jax.numpy.ndarray` (6)
        3D Spherical coordinate including time derivatives.
        rho : Radial distance in the same units of x, y, and z.
        lon : Longitude ranging from 0.0 to 360.0 degrees.
        lat : Latitude ranging from -90.0 to 90.0 degrees with 0 at the equator.
        vrho : Radial velocity in the same units as rho per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlon : Longitudinal velocity in degrees per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlat :Latitudinal velocity in degrees per arbitrary unit of time.
            (same unit of time as the x, y, and z velocities).
    """
    coords_spherical = jnp.zeros(6, dtype=jnp.float64)
    x = coords_cartesian[0]
    y = coords_cartesian[1]
    z = coords_cartesian[2]
    vx = coords_cartesian[3]
    vy = coords_cartesian[4]
    vz = coords_cartesian[5]

    rho = jnp.sqrt(x**2 + y**2 + z**2)
    lon = jnp.arctan2(y, x)
    lon = jnp.where(lon < 0.0, 2 * jnp.pi + lon, lon)
    lat = lax.cond(
        rho == 0.0,
        lambda _: 0.0,
        lambda _: jnp.arcsin(z / rho),
        None,
    )
    lat = jnp.where(
        (lat >= 3 * jnp.pi / 2) & (lat <= 2 * jnp.pi), lat - 2 * jnp.pi, lat
    )

    vrho = lax.cond(
        rho == 0.0,
        lambda _: 0.0,
        lambda _: (x * vx + y * vy + z * vz) / rho,
        None,
    )
    vlon = lax.cond(
        (x == 0.0) & (y == 0.0),
        lambda _: 0.0,
        lambda _: (vy * x - vx * y) / (x**2 + y**2),
        None,
    )
    vlat = lax.cond(
        ((x == 0.0) & (y == 0.0)) | (rho == 0.0),
        lambda _: 0.0,
        lambda _: (vz - vrho * z / rho) / jnp.sqrt(x**2 + y**2),
        None,
    )

    coords_spherical = coords_spherical.at[0].set(rho)
    coords_spherical = coords_spherical.at[1].set(jnp.degrees(lon))
    coords_spherical = coords_spherical.at[2].set(jnp.degrees(lat))
    coords_spherical = coords_spherical.at[3].set(vrho)
    coords_spherical = coords_spherical.at[4].set(jnp.degrees(vlon))
    coords_spherical = coords_spherical.at[5].set(jnp.degrees(vlat))

    return coords_spherical


# Vectorization Map: _cartesian_to_spherical
_cartesian_to_spherical_vmap = jit(
    vmap(
        _cartesian_to_spherical,
        in_axes=(0,),
    )
)


def cartesian_to_spherical(
    coords_cartesian: Union[np.ndarray, jnp.ndarray]
) -> jnp.ndarray:
    """
    Convert Cartesian coordinates to a spherical coordinates.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of distance.
        y : y-position in units of distance.
        z : z-position in units of distance.
        vx : x-velocity in the same units of x per arbitrary unit of time.
        vy : y-velocity in the same units of y per arbitrary unit of time.
        vz : z-velocity in the same units of z per arbitrary unit of time.

    Returns
    -------
    coords_spherical : ~jax.numpy.ndarray` (N, 6)
        3D Spherical coordinates including time derivatives.
        rho : Radial distance in the same units of x, y, and z.
        lon : Longitude ranging from 0.0 to 360.0 degrees.
        lat : Latitude ranging from -90.0 to 90.0 degrees with 0 at the equator.
        vrho : Radial velocity in the same units as rho per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlon : Longitudinal velocity in degrees per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlat : Latitudinal velocity in degrees per arbitrary unit of time.
            (same unit of time as the x, y, and z velocities).
    """
    coords_spherical = _cartesian_to_spherical_vmap(coords_cartesian)
    return coords_spherical


@jit
def _spherical_to_cartesian(
    coords_spherical: Union[np.ndarray, jnp.ndarray]
) -> jnp.ndarray:
    """
    Convert a single spherical coordinate to a Cartesian coordinate.

    Parameters
    ----------
    coords_spherical : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        3D Spherical coordinate including time derivatives.
        rho : Radial distance in the same units of x, y, and z.
        lon : Longitude ranging from 0.0 to 360.0 degrees.
        lat : Latitude ranging from -90.0 to 90.0 degrees with 0 at the equator.
        vrho : Radial velocity in the same units as rho per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlon : Longitudinal velocity in degrees per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlat : Latitudinal velocity in degrees per arbitrary unit of time.
            (same unit of time as the x, y, and z velocities).

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of distance.
        y : y-position in units of distance.
        z : z-position in units of distance.
        vx : x-velocity in the same units of x per arbitrary unit of time.
        vy : y-velocity in the same units of y per arbitrary unit of time.
        vz : z-velocity in the same units of z per arbitrary unit of time.
    """
    coords_cartesian = jnp.zeros(6, dtype=jnp.float64)
    rho = coords_spherical[0]
    lon = jnp.radians(coords_spherical[1])
    lat = jnp.radians(coords_spherical[2])
    vrho = coords_spherical[3]
    vlon = jnp.radians(coords_spherical[4])
    vlat = jnp.radians(coords_spherical[5])

    cos_lat = jnp.cos(lat)
    sin_lat = jnp.sin(lat)
    cos_lon = jnp.cos(lon)
    sin_lon = jnp.sin(lon)

    x = rho * cos_lat * cos_lon
    y = rho * cos_lat * sin_lon
    z = rho * sin_lat

    vx = (
        cos_lat * cos_lon * vrho
        - rho * cos_lat * sin_lon * vlon
        - rho * sin_lat * cos_lon * vlat
    )
    vy = (
        cos_lat * sin_lon * vrho
        + rho * cos_lat * cos_lon * vlon
        - rho * sin_lat * sin_lon * vlat
    )
    vz = sin_lat * vrho + rho * cos_lat * vlat

    coords_cartesian = coords_cartesian.at[0].set(x)
    coords_cartesian = coords_cartesian.at[1].set(y)
    coords_cartesian = coords_cartesian.at[2].set(z)
    coords_cartesian = coords_cartesian.at[3].set(vx)
    coords_cartesian = coords_cartesian.at[4].set(vy)
    coords_cartesian = coords_cartesian.at[5].set(vz)

    return coords_cartesian


# Vectorization Map: _spherical_to_cartesian
_spherical_to_cartesian_vmap = jit(
    vmap(
        _spherical_to_cartesian,
        in_axes=(0,),
    )
)


def spherical_to_cartesian(
    coords_spherical: Union[np.ndarray, jnp.ndarray]
) -> jnp.ndarray:
    """
    Convert spherical coordinates to Cartesian coordinates.

    Parameters
    ----------
    coords_spherical : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        3D Spherical coordinates including time derivatives.
        rho : Radial distance in the same units of x, y, and z.
        lon : Longitude ranging from 0.0 to 360.0 degrees.
        lat : Latitude ranging from -90.0 to 90.0 degrees with 0 at the equator.
        vrho : Radial velocity in the same units as rho per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlon : Longitudinal velocity in degrees per arbitrary unit of time
            (same unit of time as the x, y, and z velocities).
        vlat :Latitudinal velocity in degrees per arbitrary unit of time.
            (same unit of time as the x, y, and z velocities).

    Returns
    -------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of distance.
        y : y-position in units of distance.
        z : z-position in units of distance.
        vx : x-velocity in the same units of x per arbitrary unit of time.
        vy : y-velocity in the same units of y per arbitrary unit of time.
        vz : z-velocity in the same units of z per arbitrary unit of time.
    """
    coords_cartesian = _spherical_to_cartesian_vmap(coords_spherical)
    return coords_cartesian


@jit
def _cartesian_to_keplerian(
    coords_cartesian: Union[np.ndarray, jnp.ndarray],
    t0: float,
    mu: float,
) -> jnp.ndarray:
    """
    Convert a single Cartesian coordinate to a Keplerian coordinate.

    If the orbit is found to be circular (e = 0 +- 1e-15) then
    the argument of periapsis is set to 0. The anomalies are then accordingly
    defined with this assumption.

    If the orbit's inclination is zero or 180 degrees (i = 0 +- 1e-15 or i = 180 +- 1e-15),
    then the longitude of the ascending node is set to 0 (located in the direction of
    the reference axis).

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    t0 : float (1)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.

    Returns
    -------
    coords_keplerian : `~jax.numpy.ndarray` (13)
        13D Keplerian coordinate.
        a : semi-major axis in au.
        p : semi-latus rectum in au.
        q : periapsis distance in au.
        Q : apoapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
        nu : true anomaly in degrees.
        n : mean motion in degrees per day.
        P : period in days.
        tp : time of periapsis passage in days.

    References
    ----------
    [1] Bate, R. R; Mueller, D. D; White, J. E. (1971). Fundamentals of Astrodynamics. 1st ed.,
        Dover Publications, Inc. ISBN-13: 978-0486600611
    """
    from ..dynamics.kepler import (
        calc_mean_anomaly,
        calc_mean_motion,
        calc_periapsis_distance,
    )

    coords_keplerian = jnp.zeros(13, dtype=jnp.float64)
    r = coords_cartesian[0:3]
    v = coords_cartesian[3:6]

    r_mag = jnp.linalg.norm(r)
    v_mag = jnp.linalg.norm(v)

    sme = v_mag**2 / 2 - mu / r_mag

    # Calculate the angular momentum vector
    # Equation 2.4-1 in Bate, Mueller, & White [1]
    h = jnp.cross(r, v)
    h_mag = jnp.linalg.norm(h)

    # Calculate the vector which is perpendicular to the
    # momentum vector and the Z-axis and points towards
    # the direction of the ascending node.
    # Equation 2.4-3 in Bate, Mueller, & White [1]
    n = jnp.cross(Z_AXIS, h)
    n_mag = jnp.linalg.norm(n)

    # Calculate the eccentricity vector which lies in the orbital plane
    # and points toward periapse.
    # Equation 2.4-5 in Bate, Mueller, & White [1]
    e_vec = ((v_mag**2 - mu / r_mag) * r - (jnp.dot(r, v)) * v) / mu
    e = jnp.linalg.norm(e_vec)

    # Calculate the semi-latus rectum
    p = h_mag**2 / mu

    # Calculate the inclination
    # Equation 2.4-7 in Bate, Mueller, & White [1]
    i = jnp.arccos(h[2] / h_mag)

    # Calculate the longitude of the ascending node
    # Equation 2.4-8 in Bate, Mueller, & White [1]
    raan = jnp.arccos(n[0] / n_mag)
    raan = jnp.where(n[1] < 0, 2 * jnp.pi - raan, raan)
    # In certain conventions when the orbit is zero inclined or 180 inclined
    # the ascending node is set to 0 as opposed to being undefined. This is what
    # SPICE does so we will do the same.
    raan = jnp.where(
        (i < FLOAT_TOLERANCE) | (jnp.abs(i - 2 * jnp.pi) < FLOAT_TOLERANCE), 0.0, raan
    )

    # Calculate the argument of periapsis
    # Equation 2.4-9 in Bate, Mueller, & White [1]
    ap = jnp.arccos(jnp.dot(n, e_vec) / (n_mag * e))
    # Adopt convention that if the orbit is circular the argument of
    # periapsis is set to 0
    ap = jnp.where(e_vec[2] < 0, 2 * jnp.pi - ap, ap)
    ap = jnp.where(jnp.abs(e) < FLOAT_TOLERANCE, 0.0, ap)

    # Calculate true anomaly (undefined for
    # circular orbits)
    # Equation 2.4-10 in Bate, Mueller, & White [1]
    nu = jnp.arccos(jnp.dot(e_vec, r) / (e * r_mag))
    nu = jnp.where(jnp.dot(r, v) < 0, 2 * jnp.pi - nu, nu)
    # nu = jnp.where(jnp.abs(e) < FLOAT_TOLERANCE, jnp.nan, nu)

    # Calculate the semi-major axis (undefined for parabolic
    # orbits)
    a = jnp.where(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        jnp.nan,
        mu / (-2 * sme),
    )

    # Calculate the periapsis distance
    q = jnp.where(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        p / 2,
        calc_periapsis_distance(a, e),
    )

    # Calculate the apoapsis distance (infinite for
    # parabolic and hyperbolic orbits)
    Q = jnp.where(e < 1.0, a * (1 + e), jnp.inf)

    # Calculate the mean anomaly
    M = calc_mean_anomaly(nu, e)

    # Calculate the mean motion
    n = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda a, q: jnp.sqrt(mu / (2 * q**3)),
        lambda a, q: calc_mean_motion(a, mu),
        a,
        q,
    )

    # Calculate the orbital period which for parabolic and hyperbolic
    # orbits is infinite while for all closed orbits
    # is well defined.
    P = lax.cond(
        e < (1.0 - FLOAT_TOLERANCE), lambda n: 2 * jnp.pi / n, lambda n: jnp.inf, n
    )

    # In the case of closed orbits, if the mean anomaly is
    # greater than 180 degrees then the orbit is
    # approaching periapsis passage in which case
    # the periapsis will occur in the future
    # in less than half a period. If the mean anomaly is less
    # than 180 degrees, then the orbit is ascending from periapsis
    # passage and the most recent periapsis was in the past.
    dtp = M / n
    dtp = jnp.where((M > jnp.pi) & (e < (1.0 - FLOAT_TOLERANCE)), P - M / n, -M / n)
    tp = t0 + dtp

    coords_keplerian = coords_keplerian.at[0].set(a)
    coords_keplerian = coords_keplerian.at[1].set(p)
    coords_keplerian = coords_keplerian.at[2].set(q)
    coords_keplerian = coords_keplerian.at[3].set(Q)
    coords_keplerian = coords_keplerian.at[4].set(e)
    coords_keplerian = coords_keplerian.at[5].set(jnp.degrees(i))
    coords_keplerian = coords_keplerian.at[6].set(jnp.degrees(raan))
    coords_keplerian = coords_keplerian.at[7].set(jnp.degrees(ap))
    coords_keplerian = coords_keplerian.at[8].set(jnp.degrees(M))
    coords_keplerian = coords_keplerian.at[9].set(jnp.degrees(nu))
    coords_keplerian = coords_keplerian.at[10].set(jnp.degrees(n))
    coords_keplerian = coords_keplerian.at[11].set(P)
    coords_keplerian = coords_keplerian.at[12].set(tp)

    return coords_keplerian


# Vectorization Map: _cartesian_to_keplerian
_cartesian_to_keplerian_vmap = jit(
    vmap(
        _cartesian_to_keplerian,
        in_axes=(0, 0, 0),
    )
)


@jit
def _cartesian_to_keplerian6(
    coords_cartesian: Union[np.ndarray, jnp.ndarray],
    t0: float,
    mu: float,
) -> jnp.ndarray:
    """
    Limit conversion of Cartesian coordinates to Keplerian 6 fundamental coordinates.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    t0 : float (1)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.

    Returns
    -------
    coords_keplerian : `~jax.numpy.ndarray` (6)
        6D Keplerian coordinate.
        a : semi-major axis in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
    """
    coords_keplerian = _cartesian_to_keplerian(coords_cartesian, t0, mu)
    return coords_keplerian[jnp.array([0, 4, 5, 6, 7, 8])]


def cartesian_to_keplerian(
    coords_cartesian: Union[np.ndarray, jnp.ndarray],
    t0: Union[np.ndarray, jnp.ndarray],
    mu: Union[np.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """
    Convert Cartesian coordinates to Keplerian coordinates.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    t0 : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.

    Returns
    -------
    coords_keplerian : `~jax.numpy.ndarray` (N, 13)
        13D Keplerian coordinates.
        a : semi-major axis in au.
        p : semi-latus rectum in au.
        q : periapsis distance in au.
        Q : apoapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
        nu : true anomaly in degrees.
        n : mean motion in degrees per day.
        P : period in days.
        tp : time of periapsis passage in days.
    """
    coords_keplerian = _cartesian_to_keplerian_vmap(coords_cartesian, t0, mu)
    return coords_keplerian


@jit
def _keplerian_to_cartesian_p(
    coords_keplerian: Union[np.ndarray, jnp.ndarray],
    mu: float,
    max_iter: int = 1000,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert a single Keplerian coordinate to a Cartesian coordinate.

    Parabolic orbits (e = 1.0 +- 1e-15) with elements (a, e, i, raan, ap, M) cannot be converted
    to Cartesian orbits since their semi-major axes are by definition undefined.
    Please consider representing the orbits with Cometary elements
    and using those to convert to Cartesian. See `~adam_core.coordinates.cometary._cometary_to_cartesian`.

    Parameters
    ----------
    coords_keplerian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        6D Keplerian coordinate.
        p : semi-latus rectum in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
    mu : float
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    """
    from ..dynamics.barker import solve_barker
    from ..dynamics.kepler import solve_kepler

    coords_cartesian = jnp.zeros(6, dtype=jnp.float64)

    p = coords_keplerian[0]
    e = coords_keplerian[1]
    i = jnp.radians(coords_keplerian[2])
    raan = jnp.radians(coords_keplerian[3])
    ap = jnp.radians(coords_keplerian[4])
    M = jnp.radians(coords_keplerian[5])

    # Calculate the true anomaly
    nu = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda e_i, M_i: solve_barker(M_i),
        lambda e_i, M_i: solve_kepler(e_i, M_i, max_iter=max_iter, tol=tol),
        e,
        M,
    )

    # Calculate the perifocal rotation matrices
    r_PQW = jnp.array(
        [
            p * jnp.cos(nu) / (1 + e * jnp.cos(nu)),
            p * jnp.sin(nu) / (1 + e * jnp.cos(nu)),
            0,
        ]
    )
    v_PQW = jnp.array(
        [-jnp.sqrt(mu / p) * jnp.sin(nu), jnp.sqrt(mu / p) * (e + jnp.cos(nu)), 0]
    )

    cos_raan = jnp.cos(raan)
    sin_raan = jnp.sin(raan)
    cos_ap = jnp.cos(ap)
    sin_ap = jnp.sin(ap)
    cos_i = jnp.cos(i)
    sin_i = jnp.sin(i)

    P1 = jnp.array(
        [
            [cos_ap, -sin_ap, 0.0],
            [sin_ap, cos_ap, 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=jnp.float64,
    )

    P2 = jnp.array(
        [
            [1.0, 0.0, 0.0],
            [0.0, cos_i, -sin_i],
            [0.0, sin_i, cos_i],
        ],
        dtype=jnp.float64,
    )

    P3 = jnp.array(
        [
            [cos_raan, -sin_raan, 0.0],
            [sin_raan, cos_raan, 0.0],
            [0.0, 0.0, 1.0],
        ],
        dtype=jnp.float64,
    )

    rotation_matrix = P3 @ P2 @ P1
    r = rotation_matrix @ r_PQW
    v = rotation_matrix @ v_PQW

    coords_cartesian = coords_cartesian.at[0].set(r[0])
    coords_cartesian = coords_cartesian.at[1].set(r[1])
    coords_cartesian = coords_cartesian.at[2].set(r[2])
    coords_cartesian = coords_cartesian.at[3].set(v[0])
    coords_cartesian = coords_cartesian.at[4].set(v[1])
    coords_cartesian = coords_cartesian.at[5].set(v[2])

    return coords_cartesian


# Vectorization Map: _keplerian_to_cartesian_p
_keplerian_to_cartesian_p_vmap = jit(
    vmap(
        _keplerian_to_cartesian_p,
        in_axes=(0, 0, None, None),
    )
)


@jit
def _keplerian_to_cartesian_a(
    coords_keplerian: Union[np.ndarray, jnp.ndarray],
    mu: float,
    max_iter: int = 1000,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert a single Keplerian coordinate to a Cartesian coordinate.

    Parabolic orbits (e = 1.0 +- 1e-15) with elements (a, e, i, raan, ap, M) cannot be converted
    to Cartesian orbits since their semi-major axes are by definition undefined.
    Please consider representing the orbits with Cometary elements
    and using those to convert to Cartesian. See `~adam_core.coordinates.cometary._cometary_to_cartesian`.

    Parameters
    ----------
    coords_keplerian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        6D Keplerian coordinate.
        a : semi-major axis in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    """
    coords_keplerian_p = jnp.zeros(6, dtype=jnp.float64)
    coords_keplerian_p = coords_keplerian_p.at[1].set(coords_keplerian[1])
    coords_keplerian_p = coords_keplerian_p.at[2].set(coords_keplerian[2])
    coords_keplerian_p = coords_keplerian_p.at[3].set(coords_keplerian[3])
    coords_keplerian_p = coords_keplerian_p.at[4].set(coords_keplerian[4])
    coords_keplerian_p = coords_keplerian_p.at[5].set(coords_keplerian[5])

    # Calculate the semi-latus rectum
    a = coords_keplerian[0]
    e = coords_keplerian[1]
    p = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda a, e: jnp.nan,  # 2 * q (not enough information present)
        lambda a, e: a * (1 - e**2),
        a,
        e,
    )
    coords_keplerian_p = coords_keplerian_p.at[0].set(p)

    coords_cartesian = _keplerian_to_cartesian_p(
        coords_keplerian_p,
        mu=mu,
        max_iter=max_iter,
        tol=tol,
    )
    return coords_cartesian


# Vectorization Map: _keplerian_to_cartesian_a
_keplerian_to_cartesian_a_vmap = jit(
    vmap(
        _keplerian_to_cartesian_a,
        in_axes=(0, 0, None, None),
    )
)


@jit
def _keplerian_to_cartesian_q(
    coords_keplerian: Union[np.ndarray, jnp.ndarray],
    mu: float,
    max_iter: int = 1000,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert a single Keplerian coordinate to a Cartesian coordinate.

    Parabolic orbits (e = 1.0 +- 1e-15) with elements (a, e, i, raan, ap, M) cannot be converted
    to Cartesian orbits since their semi-major axes are by definition undefined.
    Please consider representing the orbits with Cometary elements
    and using those to convert to Cartesian. See `~adam_core.coordinates.cometary._cometary_to_cartesian`.

    Parameters
    ----------
    coords_keplerian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        6D Keplerian coordinate.
        q : periapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    """
    coords_keplerian_p = jnp.zeros(6, dtype=jnp.float64)
    coords_keplerian_p = coords_keplerian_p.at[1].set(coords_keplerian[1])
    coords_keplerian_p = coords_keplerian_p.at[2].set(coords_keplerian[2])
    coords_keplerian_p = coords_keplerian_p.at[3].set(coords_keplerian[3])
    coords_keplerian_p = coords_keplerian_p.at[4].set(coords_keplerian[4])
    coords_keplerian_p = coords_keplerian_p.at[5].set(coords_keplerian[5])

    # Calculate the semi-latus rectum
    q = coords_keplerian[0]
    e = coords_keplerian[1]
    p = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda e, q: 2 * q,
        lambda e, q: q / (1 - e) * (1 - e**2),  # a = q / (1 - e), p = a / (1 - e**2)
        e,
        q,
    )
    coords_keplerian_p = coords_keplerian_p.at[0].set(p)

    coords_cartesian = _keplerian_to_cartesian_p(
        coords_keplerian_p,
        mu=mu,
        max_iter=max_iter,
        tol=tol,
    )
    return coords_cartesian


# Vectorization Map: _keplerian_to_cartesian_q
_keplerian_to_cartesian_q_vmap = jit(
    vmap(
        _keplerian_to_cartesian_q,
        in_axes=(0, 0, None, None),
    )
)


def keplerian_to_cartesian(
    coords_keplerian: Union[np.ndarray, jnp.ndarray],
    mu: Union[np.ndarray, jnp.ndarray],
    max_iter: int = 100,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert Keplerian coordinates to Cartesian coordinates.

    Parabolic orbits (e = 1.0 +- 1e-15) with elements (a, e, i, raan, ap, M) cannot be converted
    to Cartesian orbits since their semi-major axes are by definition undefined.
    Please consider representing these orbits with Cometary elements
    and using those to convert to Cartesian. See `~adam_core.coordinates.cometary.cometary_to_cartesian`.

    Parameters
    ----------
    coords_keplerian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        6D Keplerian coordinate.
        a : semi-major axis in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        M : mean anomaly in degrees.
    mu : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.

    Raises
    ------
    ValueError: When semi-major axis is less than 0 for elliptical orbits or when
        semi-major axis is greater than 0 for hyperbolic orbits.
    """
    a = coords_keplerian[:, 0]
    e = coords_keplerian[:, 1]

    parabolic = np.where((e < (1.0 + FLOAT_TOLERANCE)) & (e > (1.0 - FLOAT_TOLERANCE)))[
        0
    ]
    if len(parabolic) > 0:
        msg = (
            "Parabolic orbits (e = 1.0 +- 1e-15) are best represented using Cometary coordinates.\n"
            "Conversion to Cartesian coordinates will not yield correct results as semi-major axis\n"
            "for parabolic orbits is undefined."
        )
        logger.critical(msg)

    hyperbolic_invalid = np.where((e > (1.0 + FLOAT_TOLERANCE)) & (a > 0))[0]
    if len(hyperbolic_invalid) > 0:
        err = (
            "Semi-major axis (a) for hyperbolic orbits (e > 1 + 1e-15) should be negative. "
            f"Instead found a = {a[hyperbolic_invalid][0]} with e = {e[hyperbolic_invalid][0]}."
        )
        raise ValueError(err)

    elliptical_invalid = np.where((e < (1.0 - FLOAT_TOLERANCE)) & (a < 0))[0]
    if len(elliptical_invalid) > 0:
        err = (
            "Semi-major axis (a) for elliptical orbits (e < 1 - 1e-15) should be positive. "
            f"Instead found a = {a[elliptical_invalid][0]} with e = {e[elliptical_invalid][0]}."
        )
        raise ValueError(err)

    coords_cartesian = _keplerian_to_cartesian_a_vmap(
        coords_keplerian, mu, max_iter, tol
    )
    return coords_cartesian


@jit
def _cartesian_to_cometary(
    coords_cartesian: Union[np.ndarray, jnp.ndarray],
    t0: float,
    mu: float,
) -> jnp.ndarray:
    """
    Convert Cartesian coordinates to Cometary coordinates.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    t0 : float (1)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : float (1)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.

    Returns
    -------
    coords_cometary : `~jax.numpy.ndarray` (6)
        6D Cometary coordinate.
        q : periapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        tp : time of periapse passage in days.
    """
    coords_cometary = _cartesian_to_keplerian(coords_cartesian, t0, mu=mu)
    return coords_cometary[jnp.array([2, 4, 5, 6, 7, 12])]


# Vectorization Map: _cartesian_to_cometary
_cartesian_to_cometary_vmap = jit(
    vmap(
        _cartesian_to_cometary,
        in_axes=(0, 0, 0),
    )
)


def cartesian_to_cometary(
    coords_cartesian: Union[np.ndarray, jnp.ndarray],
    t0: Union[np.ndarray, jnp.ndarray],
    mu: Union[np.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """
    Convert Cartesian coordinates to Keplerian coordinates.

    Parameters
    ----------
    coords_cartesian : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    t0 : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.

    Returns
    -------
    coords_cometary : `~jax.numpy.ndarray` (N, 6)
        6D Cometary coordinates.
        q : periapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        tp : time of periapse passage in days.
    """
    coords_cometary = _cartesian_to_cometary_vmap(coords_cartesian, t0, mu)
    return coords_cometary


@jit
def _cometary_to_cartesian(
    coords_cometary: Union[np.ndarray, jnp.ndarray],
    t0: float,
    mu: float,
    max_iter: int = 100,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert a single Cometary coordinate to a Cartesian coordinate.

    Parameters
    ----------
    coords_cometary : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (6)
        6D Cometary coordinate.
        q : periapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        tp : time of periapse passage in days.
    t0 : float (1)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : float
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (6)
        3D Cartesian coordinate including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    """
    coords_keplerian = jnp.zeros(6, dtype=jnp.float64)

    q = coords_cometary[0]
    e = coords_cometary[1]
    i = coords_cometary[2]
    raan = coords_cometary[3]
    ap = coords_cometary[4]
    tp = coords_cometary[5]

    # Calculate the semi-major axis from the periapsis distance
    # The semi-major axis for parabolic orbits is undefined
    a = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda e, q: jnp.nan,
        lambda e, q: q / (1 - e),
        e,
        q,
    )

    # Calculate the mean motion
    n = lax.cond(
        (e > (1.0 - FLOAT_TOLERANCE)) & (e < (1.0 + FLOAT_TOLERANCE)),
        lambda a, q: jnp.sqrt(mu / (2 * q**3)),
        lambda a, q: jnp.sqrt(mu / jnp.abs(a) ** 3),
        a,
        q,
    )

    # Calculate the orbital period which for parabolic and hyperbolic
    # orbits is infinite while for all closed orbits
    # is well defined.
    # P = lax.cond(
    #    e < (1.0 - FLOAT_TOLERANCE), lambda n: 2 * jnp.pi / n, lambda n: jnp.inf, n
    # )

    # Calculate the mean anomaly
    dtp = tp - t0
    M = jnp.where(dtp > 0, 2 * jnp.pi - dtp * n, -dtp * n)

    coords_keplerian = coords_keplerian.at[0].set(q)
    coords_keplerian = coords_keplerian.at[1].set(e)
    coords_keplerian = coords_keplerian.at[2].set(i)
    coords_keplerian = coords_keplerian.at[3].set(raan)
    coords_keplerian = coords_keplerian.at[4].set(ap)
    coords_keplerian = coords_keplerian.at[5].set(jnp.degrees(M))

    coords_cartesian = _keplerian_to_cartesian_q(
        coords_keplerian, mu=mu, max_iter=max_iter, tol=tol
    )

    return coords_cartesian


# Vectorization Map: _cometary_to_cartesian
_cometary_to_cartesian_vmap = jit(
    vmap(
        _cometary_to_cartesian,
        in_axes=(0, 0, 0, None, None),
    )
)


def cometary_to_cartesian(
    coords_cometary: Union[np.ndarray, jnp.ndarray],
    t0: Union[np.ndarray, jnp.ndarray],
    mu: Union[np.ndarray, jnp.ndarray],
    max_iter: int = 100,
    tol: float = 1e-15,
) -> jnp.ndarray:
    """
    Convert Cometary coordinates to Cartesian coordinates.

    Parameters
    ----------
    coords_cometary : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N, 6)
        6D Cometary coordinate.
        q : periapsis distance in au.
        e : eccentricity.
        i : inclination in degrees.
        raan : Right ascension (longitude) of the ascending node in degrees.
        ap : argument of periapsis in degrees.
        tp : time of periapse passage in days.
    t0 : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Epoch at which cometary elements are defined in MJD TDB.
    mu : {`~numpy.ndarray`, `~jax.numpy.ndarray`} (N)
        Gravitational parameter (GM) of the attracting body in units of
        au**3 / d**2.
    max_iter : int, optional
        Maximum number of iterations over which to converge. If number of iterations is
        exceeded, will use the value of the relevant anomaly at the last iteration.
    tol : float, optional
        Numerical tolerance to which to compute anomalies using the Newtown-Raphson
        method.

    Returns
    -------
    coords_cartesian : `~jax.numpy.ndarray` (N, 6)
        3D Cartesian coordinates including time derivatives.
        x : x-position in units of au.
        y : y-position in units of au.
        z : z-position in units of au.
        vx : x-velocity in units of au per day.
        vy : y-velocity in units of au per day.
        vz : z-velocity in units of au per day.
    """
    coords_cartesian = _cometary_to_cartesian_vmap(
        coords_cometary, t0, mu, max_iter, tol
    )
    return coords_cartesian


def cartesian_to_origin(
    coords: CartesianCoordinates, origin: OriginCodes
) -> "CartesianCoordinates":
    """
    Translate coordinates to a different origin.

    Parameters
    ----------
    coords : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        Cartesian coordinates and optionally their covariances.
    origin : `~adam_core.coordinates.origin.OriginCodes`
        Desired origin. Input origins may be either `~adam_core.coordinates.origin.OriginCodes`
        or a str of an observatory code, but the output origin (this kwarg) should always be an
        `~adam_core.coordinates.origin.OriginCodes`. If you are looking to generate ephemerides
        for an observatory, please use a `~adam_core.propagator.propagator.Propagator` instead.

    Returns
    -------
    CartesianCoordinates : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        Translated Cartesian coordinates and their covariances.

    Raises
    ------
    ValueError
        If origin is not a `~adam_core.coordinates.origin.OriginCodes` object or
        a str of an observatory code.
    """
    from ..observers.state import OBSERVATORY_CODES, get_observer_state
    from ..utils.spice import get_perturber_state

    unique_origins = coords.origin.code.unique()
    vectors = np.empty(coords.values.shape, dtype=np.float64)
    times = coords.time

    for origin_in in unique_origins:

        mask = pc.equal(coords.origin.code, origin_in).to_numpy(zero_copy_only=False)

        origin_in_str = origin_in.as_py()
        # Could use try / except block here but this is more explicit
        if origin_in_str in OriginCodes.__members__:

            vectors[mask] = get_perturber_state(
                OriginCodes[origin_in_str],
                times.apply_mask(mask),
                frame=coords.frame,
                origin=origin,
            ).values

        elif origin_in_str in OBSERVATORY_CODES:

            vectors[mask] = get_observer_state(
                origin_in_str,
                times.apply_mask(mask),
                frame=coords.frame,
                origin=origin,
            ).values

        else:
            raise ValueError("Unsupported origin: {}".format(origin_in_str))

    return coords.translate(vectors, origin.name)


def cartesian_to_frame(
    coords: CartesianCoordinates, frame: Literal["ecliptic", "equatorial"]
) -> "CartesianCoordinates":
    """
    Rotate Cartesian coordinates and their covariances to the given frame.

    Parameters
    ----------
    coords : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        Cartesian coordinates and optionally their covariances.
    frame : {'ecliptic', 'equatorial'}
        Desired reference frame of the output coordinates.

    Returns
    -------
    CartesianCoordinates : `~adam_core.coordinates.cartesian.CartesianCoordinates`
        Rotated Cartesian coordinates and their covariances.
    """
    if frame == "ecliptic" and coords.frame != "ecliptic":
        return coords.rotate(TRANSFORM_EQ2EC, "ecliptic")
    elif frame == "equatorial" and coords.frame != "equatorial":
        return coords.rotate(TRANSFORM_EC2EQ, "equatorial")
    elif frame == coords.frame:
        return coords
    else:
        err = "frame should be one of {'ecliptic', 'equatorial'}"
        raise ValueError(err)


def transform_coordinates(
    coords: types.CoordinateType,
    representation_out: Optional[type[types.CoordinateType]] = None,
    frame_out: Optional[Literal["ecliptic", "equatorial"]] = None,
    origin_out: Optional[OriginCodes] = None,
) -> types.CoordinateType:
    """
    Transform coordinates between frames ('ecliptic', 'equatorial'), origins,
    and/or representations ('cartesian', 'spherical', 'keplerian', 'cometary').

    Input coordinates may be defined from multiple origins but if origin_out is
    specified, all coordinates will be transformed to that origin.

    Parameters
    ----------
    coords:
        Coordinates to transform between representations and frames.
    representation_out:
        Desired coordinate type or representation of the output coordinates. If None,
        the output coordinates will be the same type as the input coordinates.
    frame_out : {'ecliptic', 'equatorial'}
        Desired reference frame of the output coordinates.
    origin_out : `~adam_core.coordinates.origin.OriginCodes`
        Desired origin. Input origins may be either `~adam_core.coordinates.origin.OriginCodes`
        or a str of an observatory code, but the output origin (this kwarg) should always be an
        `~adam_core.coordinates.origin.OriginCodes`. If you are looking to generate ephemerides
        for an observatory, please use a `~adam_core.propagator.propagator.Propagator` instead.

    Returns
    -------
    coords_out : `~adam_core.coordinates.Coordinates`
        Coordinates in desired output representation and frame.

    Raises
    ------
    ValueError
        If frame_in, frame_out are not one of 'equatorial', 'ecliptic'.
        If representation_in, representation_out are not one of 'cartesian',
            'spherical', 'keplerian', 'cometary'.
    TypeError
        If coords is not a `~adam_core.coordinates.Coordinates` object.
        If origin_out is not a `~adam_core.coordinates.OriginCodes` object.
    """
    # Check that coords is a thor.coordinates.Coordinates object
    if not isinstance(coords, CoordinatesClasses):
        raise TypeError("Unsupported coordinate type: {}".format(type(coords)))

    if frame_out not in {None, "equatorial", "ecliptic"}:
        raise ValueError("Unsupported frame_out: {}".format(frame_out))

    if origin_out is not None and not isinstance(origin_out, OriginCodes):
        raise TypeError("Unsupported origin_out type: {}".format(type(origin_out)))

    if representation_out is None:
        representation_out_ = coords.__class__
    else:
        representation_out_ = representation_out

    if representation_out_ not in CoordinatesClasses:
        raise ValueError(
            "Unsupported representation_out: {}".format(representation_out_)
        )

    coord_frame = coords.frame
    # Extract the origins from the input coordinates. These typically correspond
    # to the name of OriginCode enums but stored as an array of strings.
    coord_origin = coords.origin.code.to_numpy(zero_copy_only=False)

    if frame_out is None:
        frame_out = coord_frame

    # If origin out is not None, then origin_out will be an OriginCode
    # passed directly to this function. Otherwise, it will be an array of strings
    # extracted from the input coordinates.
    if origin_out is None:
        origin_out = coord_origin

    # `~adam_core.coordinates.origin.Origin` support equality checks with
    # `~adam_core.coordinates.origin.OriginCodes` so we can compare them directly.
    # If its not an OriginCodes enum then origin_out will be an array of strings which
    # also can be checked for equality.
    if type(coords) == representation_out_:
        if coord_frame == frame_out and np.all(coord_origin == origin_out):
            return coords

    if not isinstance(coords, CartesianCoordinates):
        cartesian = coords.to_cartesian()
    else:
        cartesian = coords

    # Translate coordinates to new origin (if any are different from current)
    if np.any(cartesian.origin != origin_out):
        cartesian = cartesian_to_origin(cartesian, origin_out)

    # Rotate coordinates to new frame (if different from current)
    if cartesian.frame != frame_out:
        cartesian = cartesian_to_frame(cartesian, frame_out)

    # You might think this should be 'isinstance', but no! We're
    # checking whether the input is a particular class variable, not
    # an instance of a class.
    if representation_out_ is CartesianCoordinates:
        return cartesian
    return representation_out_.from_cartesian(cartesian)



import json
import os
from collections import OrderedDict
from contextlib import contextmanager
from unittest.mock import patch

import numpy as np
import numpy.testing as npt
import pytest
from astroquery.jplsbdb import SBDB

from ..sbdb import NotFoundError, _convert_SBDB_covariances, query_sbdb


def test__convert_SBDB_covariances():
    sbdb_format = np.array(
        [
            ["e_e", "e_q", "e_tp", "e_raan", "e_ap", "e_i"],
            ["q_e", "q_q", "q_tp", "q_raan", "q_ap", "q_i"],
            ["tp_e", "tp_q", "tp_tp", "tp_raan", "tp_ap", "tp_i"],
            ["raan_e", "raan_q", "raan_tp", "raan_raan", "raan_ap", "raan_i"],
            ["ap_e", "ap_q", "ap_tp", "ap_raan", "ap_ap", "ap_i"],
            ["i_e", "i_q", "i_tp", "i_raan", "i_ap", "i_i"],
        ]
    )
    adam_core_format = np.array(
        [
            ["q_q", "q_e", "q_i", "q_raan", "q_ap", "q_tp"],
            ["e_q", "e_e", "e_i", "e_raan", "e_ap", "e_tp"],
            ["i_q", "i_e", "i_i", "i_raan", "i_ap", "i_tp"],
            ["raan_q", "raan_e", "raan_i", "raan_raan", "raan_ap", "raan_tp"],
            ["ap_q", "ap_e", "ap_i", "ap_raan", "ap_ap", "ap_tp"],
            ["tp_q", "tp_e", "tp_i", "tp_raan", "tp_ap", "tp_tp"],
        ]
    )
    sbdb_format = np.array([sbdb_format])
    adam_core_format = np.array([adam_core_format])

    cometary_covariances = _convert_SBDB_covariances(sbdb_format)
    # Some of the symmetric terms will be in reverse format: for example
    # q_e will be e_q where it should be q_e. So for places where
    # the order is reversed, let's try to flip them around and then lets
    # check for equality with the expected adam format.

    flip_mask = np.where(cometary_covariances != adam_core_format)
    for i, j, k in zip(*flip_mask):
        cometary_covariances[i, j, k] = "_".join(
            cometary_covariances[i, j, k].split("_")[::-1]
        )

    npt.assert_equal(cometary_covariances, adam_core_format)


def test_query_sbdb_for_ceres():
    with mock_sbdb_query("Ceres.json") as mock:
        result = query_sbdb(["Ceres"])
        mock.assert_called_once()

    assert len(result) == 1


def test_query_sbdb_for_2001vb():
    with mock_sbdb_query("2001VB.json") as mock:
        result = query_sbdb(["2001VB"])
        mock.assert_called_once()

    assert len(result) == 1


def test_query_sbdb_for_54509():
    # 54509 has a 7x7 covariance matrix so lets test that we
    # correctly convert it to the 6x6 format
    with mock_sbdb_query("54509.json") as mock:
        assert mock.return_value["orbit"]["covariance"]["data"].shape == (7, 7)
        assert mock.return_value["orbit"]["covariance"]["labels"] == [
            "e",
            "q",
            "tp",
            "node",
            "peri",
            "i",
            "A2",
        ]
        result = query_sbdb(["54509"])
        mock.assert_called_once()

    assert len(result) == 1
    assert result.coordinates.covariance.to_matrix().shape == (1, 6, 6)
    assert not np.isnan(result.coordinates.covariance.to_matrix()).all()


def test_query_sbdb_for_missing_value():
    with pytest.raises(NotFoundError):
        with mock_sbdb_query("missing.json"):
            query_sbdb(["missing"])


@contextmanager
def mock_sbdb_query(response_file: str):
    with patch("adam_core.orbits.query.sbdb.SBDB.query") as mock_sbdb_query:
        resp_path = os.path.join(
            os.path.dirname(__file__), "testdata", "sbdb", response_file
        )
        response_dict = json.load(open(resp_path, "r"))
        response_value = SBDB()._process_data(OrderedDict(response_dict))
        mock_sbdb_query.return_value = response_value
        yield mock_sbdb_query



import numpy as np
import pytest

from ..origin import Origin, OriginCodes, OriginGravitationalParameters


def test_origin_eq__():
    origin = Origin.from_kwargs(code=["SUN", "EARTH", "SUN"])

    # Test equality with string
    np.testing.assert_equal(origin == "SUN", np.array([True, False, True]))
    np.testing.assert_equal(origin != "SUN", np.array([False, True, False]))

    # Test equality with numpy array
    np.testing.assert_equal(
        origin == np.array(["SUN", "EARTH", "SUN"]), np.array([True, True, True])
    )
    np.testing.assert_equal(
        origin == np.array(["SUN", "EARTH", "EARTH"]), np.array([True, True, False])
    )
    np.testing.assert_equal(
        origin != np.array(["SUN", "EARTH", "EARTH"]), np.array([False, False, True])
    )
    np.testing.assert_equal(
        origin == np.array(["SUN", "SUN", "SUN"]), np.array([True, False, True])
    )
    np.testing.assert_equal(
        origin != np.array(["SUN", "SUN", "SUN"]), np.array([False, True, False])
    )

    # Test equality with Origin
    np.testing.assert_equal(
        origin == Origin.from_kwargs(code=["SUN", "SUN", "SUN"]),
        np.array([True, False, True]),
    )
    np.testing.assert_equal(
        origin != Origin.from_kwargs(code=["SUN", "SUN", "SUN"]),
        np.array([False, True, False]),
    )
    np.testing.assert_equal(
        origin == Origin.from_kwargs(code=["SUN", "EARTH", "SUN"]),
        np.array([True, True, True]),
    )
    np.testing.assert_equal(
        origin != Origin.from_kwargs(code=["SUN", "EARTH", "SUN"]),
        np.array([False, False, False]),
    )

    # Test equality with OriginCodes
    np.testing.assert_equal(origin == OriginCodes.SUN, np.array([True, False, True]))
    np.testing.assert_equal(origin != OriginCodes.SUN, np.array([False, True, False]))
    np.testing.assert_equal(origin == OriginCodes.EARTH, np.array([False, True, False]))
    np.testing.assert_equal(origin != OriginCodes.EARTH, np.array([True, False, True]))


def test_origin__eq__raises():
    # Test that an error is raised when an unsupported type is passed
    origin = Origin.from_kwargs(code=["SUN", "EARTH", "MARS"])
    with pytest.raises(TypeError):
        origin == 1


def test_origin_mu():
    # Test that the mu function returns the correct values
    origin = Origin.from_kwargs(
        code=["SUN", "MARS_BARYCENTER", "JUPITER_BARYCENTER", "SUN"]
    )

    expected = np.array(
        [
            OriginGravitationalParameters.SUN,
            OriginGravitationalParameters.MARS_BARYCENTER,
            OriginGravitationalParameters.JUPITER_BARYCENTER,
            OriginGravitationalParameters.SUN,
        ]
    )
    np.testing.assert_equal(origin.mu(), expected)


def test_origin_mu_raises():
    # Test that the mu function raises a ValueError when an unsupported code is passed
    origin = Origin.from_kwargs(code=["VESTA"])
    with pytest.raises(ValueError):
        origin.mu()



# flake8: noqa: F401
from .propagator import EphemerisMixin, EphemerisType, OrbitType, Propagator
from .utils import _iterate_chunks

__all__ = [
    "Propagator",
    "EphemerisMixin",
    "OrbitType",
    "EphemerisType",
]



import numpy as np
from astropy import units as u

from ..origin import Origin
from ..transform import cartesian_to_cometary, cometary_to_cartesian


def test_cometary_to_cartesian_elliptical(orbital_elements):
    # Test cometary_to_cartesian (vmapped) against the expected cartesian elements
    # of a series of sample orbital elements.

    # Limit to elliptical orbits
    orbital_elements = orbital_elements[orbital_elements["e"] < 1]

    cometary_elements = orbital_elements[
        ["q", "e", "incl", "Omega", "w", "tp_mjd"]
    ].values
    cartesian_elements_expected = orbital_elements[
        ["x", "y", "z", "vx", "vy", "vz"]
    ].values
    t0_mjd = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    cartesian_elements_actual = cometary_to_cartesian(
        cometary_elements, t0_mjd, origin.mu()
    )
    diff = cartesian_elements_actual - cartesian_elements_expected

    # Calculate offset in position in mm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.mm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 20 m
    np.testing.assert_array_less(
        r_diff, 10 * 1000 * 2
    )  # TODO: why does this need to be so much larger than the keplerian_to_cartesian test?
    # Assert velocities are to within 10 mm/s
    np.testing.assert_array_less(v_diff, 10)


def test_cometary_to_cartesian_hyperbolic(orbital_elements):
    # Test cometary_to_cartesian (vmapped) against the expected cartesian elements
    # of a series of sample orbital elements.

    # Limit to hyperbolic orbits
    orbital_elements = orbital_elements[orbital_elements["e"] > 1]

    cometary_elements = orbital_elements[
        ["q", "e", "incl", "Omega", "w", "tp_mjd"]
    ].values
    cartesian_elements_expected = orbital_elements[
        ["x", "y", "z", "vx", "vy", "vz"]
    ].values
    t0_mjd = orbital_elements["mjd_tdb"].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    cartesian_elements_actual = cometary_to_cartesian(
        cometary_elements, t0_mjd, origin.mu()
    )
    diff = cartesian_elements_actual - cartesian_elements_expected

    # Calculate offset in position in mm
    r_diff = np.linalg.norm(diff[:, :3], axis=1) * u.au.to(u.mm)
    # Calculate offset in velocity in mm/s
    v_diff = np.linalg.norm(diff[:, 3:], axis=1) * (u.au / u.d).to(u.mm / u.s)

    # Assert positions are to within 20 m
    np.testing.assert_array_less(
        r_diff, 10 * 1000 * 2
    )  # TODO: why does this need to be so much larger than the keplerian_to_cartesian test?
    # Assert velocities are to within 10 mm/s
    np.testing.assert_array_less(v_diff, 10)


def test_cartesian_to_cometary_elliptical(orbital_elements):
    # Test cartesian_to_cometary (vmapped) against the expected cometary elements
    # of a series of sample orbital elements.

    # Limit to elliptical orbits
    orbital_elements = orbital_elements[orbital_elements["e"] < 1]

    cometary_elements_expected = orbital_elements[
        ["q", "e", "incl", "Omega", "w", "tp_mjd"]
    ].values
    epochs = orbital_elements["mjd_tdb"].values
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    # Cartesian to cometary returns an N, 13 array containing the cometary elements and cometary
    # elements, with some additional columns.
    cometary_elements_actual = cartesian_to_cometary(
        cartesian_elements, epochs, origin.mu()
    )

    # Calculate the difference
    diff = cometary_elements_actual - cometary_elements_expected

    # Calculate offset in periapsis distance in m
    q_diff = np.abs(diff[:, 0]) * u.au.to(u.m)
    # Assert periapsis distance is to within 100 m
    np.testing.assert_array_less(q_diff, 200)  # TODO

    # Calculate offset in eccentricity
    e_diff = np.abs(diff[:, 1])
    # Assert eccentricity is to within 1e-10
    np.testing.assert_array_less(e_diff, 1e-10)

    # Calculate offset in inclination in microarcseconds
    i_diff = np.abs(diff[:, 2]) * u.degree.to(u.microarcsecond)
    # Assert inclination is to within 10 microarcseconds
    np.testing.assert_array_less(i_diff, 10)

    # Calculate offset in longitude of ascending node in microarcseconds
    raan_diff = np.abs(diff[:, 3]) * u.degree.to(u.microarcsecond)
    # Assert longitude of ascending node is to within 10 microarcseconds
    np.testing.assert_array_less(raan_diff, 10)

    # Calculate offset in argument of periapsis in microarcseconds
    ap_diff = np.abs(diff[:, 4]) * u.degree.to(u.microarcsecond)
    # Assert argument of periapsis is to within 100 microarcseconds
    np.testing.assert_array_less(ap_diff, 100)  # TODO

    # Calculate offset in time of perihelion passage in seconds
    tp_diff = np.abs(diff[:, 5]) * u.d.to(u.s)
    # Assert time of perihelion passage is to within 1 second
    np.testing.assert_array_less(tp_diff, 1)

    # TODO: All TODOs should be checked to see if they are reasonable tolerances.
    #  I'd expect the semi-major axis to be much better.


def test_cartesian_to_cometary_hyperbolic(orbital_elements):
    # Test cartesian_to_cometary (vmapped) against the expected cometary elements
    # of a series of sample orbital elements.

    # Limit to hyperbolic orbits
    orbital_elements = orbital_elements[orbital_elements["e"] > 1]

    cometary_elements_expected = orbital_elements[
        ["q", "e", "incl", "Omega", "w", "tp_mjd"]
    ].values
    epochs = orbital_elements["mjd_tdb"].values
    cartesian_elements = orbital_elements[["x", "y", "z", "vx", "vy", "vz"]].values
    origin = Origin.from_kwargs(code=["SUN"] * len(orbital_elements))

    # Cartesian to cometary returns an N, 13 array containing the cometary elements and cometary
    # elements, with some additional columns.
    cometary_elements_actual = cartesian_to_cometary(
        cartesian_elements, epochs, origin.mu()
    )

    # Calculate the difference
    diff = cometary_elements_actual - cometary_elements_expected

    # Calculate offset in periapsis distance in m
    q_diff = np.abs(diff[:, 0]) * u.au.to(u.m)
    # Assert periapsis distance is to within 100 m
    np.testing.assert_array_less(q_diff, 200)  # TODO

    # Calculate offset in eccentricity
    e_diff = np.abs(diff[:, 1])
    # Assert eccentricity is to within 1e-10
    np.testing.assert_array_less(e_diff, 1e-10)

    # Calculate offset in inclination in microarcseconds
    i_diff = np.abs(diff[:, 2]) * u.degree.to(u.microarcsecond)
    # Assert inclination is to within 10 microarcseconds
    np.testing.assert_array_less(i_diff, 10)

    # Calculate offset in longitude of ascending node in microarcseconds
    raan_diff = np.abs(diff[:, 3]) * u.degree.to(u.microarcsecond)
    # Assert longitude of ascending node is to within 10 microarcseconds
    np.testing.assert_array_less(raan_diff, 10)

    # Calculate offset in argument of periapsis in microarcseconds
    ap_diff = np.abs(diff[:, 4]) * u.degree.to(u.microarcsecond)
    # Assert argument of periapsis is to within 100 microarcseconds
    np.testing.assert_array_less(ap_diff, 100)  # TODO

    # Calculate offset in time of perihelion passage in seconds
    tp_diff = np.abs(diff[:, 5]) * u.d.to(u.s)
    # Assert time of perihelion passage is to within 1 second
    np.testing.assert_array_less(tp_diff, 1)

    # TODO: All TODOs should be checked to see if they are reasonable tolerances.
    #  I'd expect the semi-major axis to be much better.
